import numpy as np
import collections as clt
import pandas as pd

from pathlib import Path
from Bio import SeqIO

import argparse
import time
import os
import re
from gooey import Gooey, GooeyParser


def get_codons():
    bases = 'TCAG'
    return [a + b + c for a in bases for b in bases for c in bases]


def invert_codon_table(ct):
    """
    input: Codon table as dictionary with codons as keys and amino acids as values
    output: Dictionary with amino acids as keys and codons as values
    """
    inv_codon_table = clt.defaultdict(list)
    for k, v in ct.items():
        inv_codon_table[v] = inv_codon_table.get(v, [])
        inv_codon_table[v].append(k)
    return inv_codon_table


def codonify(seq):
    """
    input: a nucleotide sequence (not necessarily a string)
    output: a list of codons
    """
    seq = str(seq).upper()
    l = len(seq)
    return [seq[i:i + 3] for i in range(0, l, 3)]


def count_observed_codons(dp, cds_dict):
    codon_counts = clt.Counter()
    proteinID = np.array(dp['Protein'], dtype=str)
    peptideStartPos = np.array(dp['Protein Start'], dtype=int) - 1
    peptideEndPos = np.array(dp['Protein End'], dtype=int)
    for i in range(0, len(proteinID)):
        sequence = cds_dict[proteinID[i]]
        codon_seq = codonify(sequence.seq)

        codon_peptide = codon_seq[peptideStartPos[i]:peptideEndPos[i]]
        codon_counts.update(clt.Counter(codon_peptide))

    return codon_counts


def get_mass_substitution_dict():
    """
    output: Dictionary of amino acid (key) mass differences (value)
    """
    aa_mass = {"G": 57.02147, "A": 71.03712, "S": 87.03203, "P": 97.05277, "V": 99.06842, "T": 101.04768,
               "I": 113.08407, "L": 113.08407, "N": 114.04293, "D": 115.02695, "Q": 128.05858, "K": 128.09497,
               "E": 129.0426, "M": 131.04049, "H": 137.05891, "F": 147.06842, "R": 156.10112, "C": 160.030654,  # CamCys
               "Y": 163.0633, "W": 186.07932, "*": 71.03712}

    subs_dict_complete = {i + ' to ' + j: aa_mass[j] - aa_mass[i] for i in aa_mass for j in aa_mass if i != j}
    del subs_dict_complete['L to I']
    del subs_dict_complete['I to L']

    subs_dict_complete = {k: v for k, v in subs_dict_complete.items() if k[-1] != 'L'}
    subs_dict = clt.defaultdict(int)
    for k, v in subs_dict_complete.items():  # unifies I and L
        if k[-1] == 'I':
            subs_dict[k + '/L'] = v
        else:
            subs_dict[k] = v

    return subs_dict


def mark_danger_mods(dp, dm, mass_tol=0.005):
    aas = list('ACDEFGHIKLMNPQRSTVWY')

    dp['danger'] = False
    for mod in dm.iterrows():
        mod = mod[1]
        position = mod['position']
        site = mod['site']
        delta_m = mod['delta_m']

        mass_filter = (delta_m - (2 * mass_tol) < dp['Delta Mass']) & (dp['Delta Mass'] < delta_m + (2 * mass_tol))

        site_filter = True
        if site in aas:
            site_filter = dp.modAA.str.contains(site)

        term_filter = True
        if position == 'Protein N-term':
            term_filter = dp.is_prot_nterm
        elif position == 'Protein C-term':
            term_filter = dp.is_prot_cterm
        elif position == 'Any N-term':
            term_filter = dp.is_peptide_nterm
        elif position == 'Any C-term':
            term_filter = dp.is_peptide_cterm

        dp.loc[site_filter & term_filter & mass_filter, 'danger'] = True

    return dp


def define_near_cognate_mask(codons, aas, i_codon_table):
    """
    Create mask for mispairing. A binary dataframe indicating for each codon the AAs encoded by near-cognate codons.
    """

    def hamming(s1, s2):
        return sum(a != b for a, b in zip(s1, s2))

    mask = pd.DataFrame(data=False, index=codons, columns=list('ACDEFGHKLMNPQRSTVWY*'), dtype=float)
    for label in codons:
        near_cognates = np.array([hamming(i, label) == 1 for i in codons])
        reachable_aa = set(np.array(list(aas))[near_cognates])
        mask.loc[label] = [i in reachable_aa for i in 'ACDEFGHKLMNPQRSTVWY*']

    for label in mask.index:  # removes "near-cognates" that encodes the same AA
        for col in mask.columns:
            if label in i_codon_table[col]:
                mask.loc[label, col] = float('NaN')

    return mask.rename(columns={'L': 'I/L'})


def mark_substitutions(peptide_df, mass_diff_dict, tolerance):
    peptide_df['substitution'] = False
    for i in sorted(mass_diff_dict.keys()):
        delta_m = mass_diff_dict[i]
        original_aa = i[0]
        peptide_df.loc[(peptide_df['Delta Mass'] > (delta_m - tolerance)) &
                       (peptide_df['Delta Mass'] < (delta_m + tolerance)) &
                       (peptide_df['modAA'] == original_aa) & ~peptide_df['danger'], 'substitution'] = i
    return peptide_df


def peptide_in_proteome(peptide, proteome):
    return any([id for id, seq in proteome.items() if peptide in str(seq.seq)])


def is_valid_file(p, arg):
    if not os.path.exists(arg):
        p.error("The file %s does not exist!" % arg)
    else:
        return arg


def prep_folder(arg):
    if not os.path.exists(arg):
        os.mkdir(arg)

    return arg


def prepare(psm, dm_mass_tol, decoy, cds_names):
    psm = psm.loc[psm['Is Unique'], :].copy()
    psm['is_decoy'] = psm['Protein'].map(lambda p: re.search(decoy, p) is not None)
    psm['is_contaminant'] = psm['Protein'].map(lambda p: p not in cds_names)
    psm.query("not is_decoy and not is_contaminant", inplace=True)
    psm['Delta Mass'] = psm['Delta Mass'].astype(float)
    psm['zero_shift_peptide'] = psm['Delta Mass'].map(lambda dm: (dm < dm_mass_tol) and (dm > -dm_mass_tol))
    psm['is_prot_nterm'] = psm['Prev AA'].map(lambda x: x == '-')
    psm['is_prot_cterm'] = psm['Next AA'].map(lambda x: x == '-')

    psm.loc[psm['MSFragger Localization'].isna(), 'MSFragger Localization'] = ''
    psm['MSFragger Localization'] = psm['MSFragger Localization'].astype(str)

    psm['n_matched_pos'] = psm['MSFragger Localization'].map(lambda loc: len(re.findall("[a-z]", loc)))
    psm.query("n_matched_pos <= 1", inplace=True)
    psm = psm.loc[~((psm['n_matched_pos'] == 1) & (psm["Delta Mass"] < dm_mass_tol) &
                    (psm["Delta Mass"] > -dm_mass_tol)), :]

    # psm.query(f" not (n_matched_pos == 1 and (`Delta Mass` < {dm_mass_tol}) and (`Delta Mass` > {-dm_mass_tol}))", inplace=True)

    def localize_mod(loc):
        x = re.search("[a-z]", loc)
        return x.span()[0] if x is not None else -1

    psm['mod_loc_in_peptide'] = psm['MSFragger Localization'].map(localize_mod)

    def get_mod_aa(loc):
        x = re.findall("[a-z]", loc)
        return x[0].upper() if len(x) > 0 else ''

    psm['modAA'] = psm['MSFragger Localization'].map(get_mod_aa)
    psm['Protein Start'] = psm['Protein Start'].astype(int)
    psm['mod_loc_in_protein'] = psm['mod_loc_in_peptide'] + psm['Protein Start']
    psm['is_peptide_nterm'] = psm['mod_loc_in_peptide'] == 0
    psm['is_peptide_cterm'] = (psm['mod_loc_in_peptide'] + 1) == psm['Peptide Length']
    psm['raw_file'] = [s[0:re.search("\.\d+\.\d+\.\d\Z", s).span()[0]] for s in psm.index]

    return psm


def retain_subs_with_basepeptide(psm):
    subs = psm.query('substitution != False and danger == False').copy()
    base_peptide_df = psm.query('substitution == False and danger == False and zero_shift_peptide')
    print(f'\tidentified {base_peptide_df.shape[0]} base peptides')

    def is_in(x, base_df):
        return base_df.loc[(base_df["Protein"] == x.Protein) & (base_df["raw_file"] == x.raw_file) &
                           (base_df["Protein Start"] < x.mod_loc_in_protein) &
                           (base_df["Protein End"] > x.mod_loc_in_protein), :].shape[0] > 0

        # return base_df.query(f"Protein == '{x.Protein}' and raw_file == '{x.raw_file}' and "
        #                     f"`Protein Start` < {x.mod_loc_in_protein} and "
        #                     f"`Protein End` > {x.mod_loc_in_protein}").shape[0] > 0

    subs['base_peptide_exists'] = subs.apply(lambda row: is_in(row, base_peptide_df), axis=1)
    return subs.query('base_peptide_exists')


def fetch_substituted_codon(row, cds_dict):
    mod_loc = row.mod_loc_in_protein
    prot = row.Protein

    codon = 'NaN'
    if (not row.is_decoy) & (prot in cds_dict.keys()):
        sequence = cds_dict[prot]
        codon_seq = codonify(sequence.seq)
        codon = codon_seq[mod_loc - 1]

    return codon


def run(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", dest="fasta", required=True,
                        help="search database as codon fasta file", metavar="FILE",
                        type=lambda x: is_valid_file(parser, x))
    parser.add_argument("-psm", dest="psm", required=True,
                        help="psm.tsv produced by philisopher", metavar="FILE")
    parser.add_argument("-o", dest="output_folder", required=True,
                        help="Output folder", metavar="FILE",
                        type=lambda x: prep_folder(x))
    parser.add_argument("-decoy", dest="decoy", required=False,
                        help="decoy prefix (default: rev_)", default="rev_")
    parser.add_argument("-p", dest="prefix", required=False,
                        help="prefix for experiment naming", default='substitutions')
    parser.add_argument("-tol", dest="tol", required=False,
                        help="m/z tolerance, used to filter DP–BP couples that resemble substitutions "
                             "and exclude pairs that resemble known PTM", default=0.005, type=float)
    parser.add_argument("-TEST", dest="test", required=False,
                        help="For testing only: use only <X> files", default=-1)
    args = parser.parse_args(args)

    print("Setup")
    tic_total = time.perf_counter()

    ntest_files = int(args.test)
    TESTING = ntest_files > -1

    print("\n\n\n")
    print("--------------------------------------------------------------------")
    print(f"fasta database file: {args.fasta}")
    print(f"psm file: {args.psm}")
    print(f"decoy prefix: {args.decoy}")
    print(f"Output folder: {args.output_folder}")
    print(f"Output file prefix: {args.prefix}")
    print(f"M/Z tolerance: {args.tol}")
    print("--------------------------------------------------------------------")

    # to be set via command line arguments
    delta_mass_tolerance = args.tol  # 0.005
    cds_file = args.fasta  #
    psm_file = args.psm  # 'input/PXD025934_fragpipe/pepXML'
    output_folder = args.output_folder  # 'output/PXD025934_fragpipe/'
    decoy_prefix = args.decoy
    output_prefix = args.prefix

    ms_diff_dict = get_mass_substitution_dict()
    codons = get_codons()
    amino_acids = 'FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG'  # corresponds to codons
    codon_table = dict(zip(codons, amino_acids))
    inverted_codon_table = invert_codon_table(codon_table)
    inverted_codon_table['L'] = inverted_codon_table['L'] + inverted_codon_table['I']

    print("Reading sequences")
    tic = time.perf_counter()
    cds_dict = SeqIO.to_dict(SeqIO.parse(cds_file, 'fasta'))
    toc = time.perf_counter()
    print(f"\tfinished in {toc - tic:0.4f} seconds")
    print(f"\t# of sequences in database: {len(cds_dict)}")

    print('Reading and preparing PSM file')
    tic = time.perf_counter()
    psm_df = pd.read_csv(psm_file, sep='\t', index_col=0)
    psm_df = prepare(psm_df, dm_mass_tol=delta_mass_tolerance, decoy=decoy_prefix, cds_names=cds_dict.keys())
    toc = time.perf_counter()
    print(f"\tfinished in {toc - tic:0.4f} seconds")

    print('Finishing up substitution detection')
    tic = time.perf_counter()
    p = Path(__file__).with_name('danger_mods.csv')
    danger_mods = pd.read_csv(p)  # pd.read_csv('atplab_workflow/danger_mods.csv') #

    print("\tMark dangerous PTMs")
    psm_df = mark_danger_mods(psm_df, danger_mods, mass_tol=delta_mass_tolerance)
    psm_df = mark_substitutions(psm_df, ms_diff_dict, delta_mass_tolerance)
    psm_subs = retain_subs_with_basepeptide(psm_df)
    # psm_subs['found_elsewhere'] = psm_subs['modified_peptide'].map(lambda x: peptide_in_proteome(x, cds_dict))

    near_cognate_mask = define_near_cognate_mask(codons, amino_acids, inverted_codon_table)

    psm_subs['codon'] = psm_subs.apply(lambda row: fetch_substituted_codon(row, cds_dict), axis=1)
    psm_subs['origin'] = psm_subs['substitution'].map(lambda x: x.split(' ')[0])
    psm_subs['destination'] = psm_subs['substitution'].map(lambda x: x.split(' ')[2])
    psm_subs['near_cognate'] = psm_subs.apply(lambda row: near_cognate_mask.loc[row['codon'], row['destination']],
                                              axis=1)

    psm_subs = psm_subs[['Protein', 'Peptide', 'raw_file', 'MSFragger Localization', 'Prev AA', 'Next AA',
                         'PeptideProphet Probability', 'Delta Mass', 'Protein Start', 'Protein End',
                         'mod_loc_in_protein', 'codon', 'origin', 'destination', 'substitution', 'near_cognate']]

    psm_subs.rename(columns={'raw_file': 'Rawfile', 'MSFragger Localization': 'Modified_Peptide', 'Prev AA': 'Prev_AA',
                             'Next AA': 'Next_AA', 'PeptideProphet Probability': 'PeptideProphet_Probability',
                             'Delta Mass': 'Delta_Mass', 'Protein Start': 'Protein_Start', 'Protein End': 'Protein_End',
                             'mod_loc_in_protein': 'Localization_in_Protein'}, inplace=True)

    psm_subs = psm_subs.astype({'Protein': str, 'Peptide': str, 'Rawfile': str, 'Modified_Peptide': str, 'Prev_AA': str,
                                'Next_AA': str, 'PeptideProphet_Probability': float, 'Delta_Mass': float,
                                'Protein_Start': int, 'Protein_End': int, 'Localization_in_Protein': int, 'codon': str,
                                'origin': str, 'destination': str, 'substitution': str, 'near_cognate': bool})

    toc = time.perf_counter()
    print(f"\tfinished in {toc - tic:0.4f} seconds")
    print(f"\tDetected {psm_subs.shape[0]} substitutions")

    print('Calculating Protein and Peptide error rates')
    tic = time.perf_counter()
    # peptide count
    psm_pc = pd.DataFrame.from_dict(clt.Counter(psm_df['Protein']), orient='index')
    psm_zs_pc = pd.DataFrame.from_dict(clt.Counter(psm_df.loc[psm_df['zero_shift_peptide'], 'Protein']), orient='index')
    psm_subs_pc = pd.DataFrame.from_dict(clt.Counter(psm_subs['Protein']), orient='index')
    psm_pc = psm_pc.join(psm_zs_pc, lsuffix='Total', rsuffix='zero_shift')
    psm_pc = psm_pc.join(psm_subs_pc, lsuffix='Total', rsuffix='subs')
    psm_pc['Error_rate'] = psm_pc[0] / psm_pc['0Total']
    psm_pc['ZS_Error_rate'] = psm_pc[0] / psm_pc['0zero_shift']
    psm_pc = psm_pc.reset_index()
    psm_pc.fillna(0, inplace=True)
    psm_pc.columns = ['Protein', 'Total', 'No_Mass_Shift', 'Erroneous', 'Total_Error_Rate', 'No_Mass_Shift_Error_Rate']

    # codon counting
    base_codon_count = count_observed_codons(psm_df.query('zero_shift_peptide'), cds_dict)
    basepeptide_codon_count = pd.DataFrame.from_dict(base_codon_count, orient='index')
    subs_codon_count = pd.DataFrame.from_dict(clt.Counter(psm_subs['codon']), orient='index')
    total_codon_count = basepeptide_codon_count.join(subs_codon_count, lsuffix='base', rsuffix='substitution')
    total_codon_count['total'] = total_codon_count['0base'] + total_codon_count['0substitution']
    total_codon_count['error_rate'] = total_codon_count['0substitution'] / total_codon_count['total']
    total_codon_count = total_codon_count.reset_index()
    total_codon_count.fillna(0, inplace=True)
    total_codon_count.columns = ['codon', 'base_count', 'error_count', 'total_count', 'detection_rate']
    psm_cc = total_codon_count.astype({'codon': str, 'base_count': int, 'error_count': int, 'total_count': int,
                                       'detection_rate': float})
    toc = time.perf_counter()
    print(f"\tfinished in {toc - tic:0.4f} seconds")

    psm_subs.to_csv(os.path.join(output_folder, output_prefix + '_substitution_errors.csv'))
    psm_cc.to_csv(os.path.join(output_folder, output_prefix + '_codon_counts.csv'))
    psm_pc.to_csv(os.path.join(output_folder, output_prefix + '_peptide_counts.csv'))

    toc_total = time.perf_counter()
    print(f"Analysis finished in {toc_total - tic_total:0.4f} seconds")


@Gooey(program_name='eTEL', default_size=(610, 580))
def main(args):
    parser = GooeyParser(description="eTEL detects the empirical translation error landscape.")
    parser.add_argument("-f", dest="fasta", required=True,
                        help="search database as codon fasta file", metavar="FILE",
                        type=lambda x: is_valid_file(parser, x), widget="FileChooser")
    parser.add_argument("-psm", dest="psm", required=True,
                        help="psm.tsv produced by philisopher", metavar="FILE", widget="FileChooser")
    parser.add_argument("-o", dest="output_folder", required=True,
                        help="Output folder", metavar="FILE",
                        type=lambda x: prep_folder(x), widget="DirChooser")
    parser.add_argument("-decoy", dest="decoy", required=False,
                        help="decoy prefix (default: rev_)", default="rev_")
    parser.add_argument("-p", dest="prefix", required=False,
                        help="prefix for experiment naming", default='substitutions')
    parser.add_argument("-tol", dest="tol", required=False,
                        help="m/z tolerance, used to filter DP–BP couples that resemble substitutions "
                             "and exclude pairs that resemble known PTM", default=0.005)
    parser.add_argument("-TEST", dest="test", required=False,
                        help="For testing only: use only <X> files", default=-1)
    args = parser.parse_args()
    run(args)


if __name__ == '__main__':
    import sys

    main(sys.argv[1:])



"""
deTEL python.

deTEL python package
"""
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

__version__ = "0.1.0"
__author__ = 'Cedric Landerer'
__credits__ = 'Agnes Toth-Petroczy Group'



"""
Most routines are taken from the workflow created by Ernest Mordret (Mordret et al. 2019, Cell)
The functions were cleaned up to eliminate global variables and to make the workflow more command line and git friendly.
"""
from itertools import groupby
from operator import itemgetter
from numba import njit

import collections as clt
import pandas as pd
import numpy as np

import re


def create_suffix_array(text, _step=16):
    """Analyze all common strings in the text.

    Short substrings of the length _step a are first pre-sorted. The are the
    results repeatedly merged so that the garanteed number of compared
    characters bytes is doubled in every iteration until all substrings are
    sorted exactly.

    Arguments:
        text:  The text to be analyzed.
        _step: Is only for optimization and testing. It is the optimal length
               of substrings used for initial pre-sorting. The bigger value is
               faster if there is enough memory. Memory requirements are
               approximately (estimate for 32 bit Python 3.3):
                   len(text) * (29 + (_size + 20 if _size > 2 else 0)) + 1MB

    Return value:      (tuple)
      (sa, rsa, lcp)
        sa:  Suffix array                  for i in range(1, size):
               assert text[sa[i-1]:] < text[sa[i]:]
        rsa: Reverse suffix array          for i in range(size):
               assert rsa[sa[i]] == i
        lcp: Longest common prefix         for i in range(1, size):
               assert text[sa[i-1]:sa[i-1]+lcp[i]] == text[sa[i]:sa[i]+lcp[i]]
               if sa[i-1] + lcp[i] < len(text):
                   assert text[sa[i-1] + lcp[i]] < text[sa[i] + lcp[i]]

    suffix_array(text='banana')
    ([5, 3, 1, 0, 4, 2], [3, 2, 5, 1, 4, 0], [0, 1, 3, 0, 0, 2])

    Explanation: 'a' < 'ana' < 'anana' < 'banana' < 'na' < 'nana'
    The Longest Common String is 'ana': lcp[2] == 3 == len('ana')
    It is between  tx[sa[1]:] == 'ana' < 'anana' == tx[sa[2]:]
    """
    tx = text
    size = len(tx)
    step = min(max(_step, 1), len(tx))
    sa = list(range(len(tx)))
    sa.sort(key=lambda i: tx[i:i + step])
    grpstart = size * [False] + [True]  # a boolean map for iteration speedup.
    # It helps to skip yet resolved values. The last value True is a sentinel.
    rsa = size * [None]
    stgrp, igrp = '', 0
    for i, pos in enumerate(sa):
        st = tx[pos:pos + step]
        if st != stgrp:
            grpstart[igrp] = (igrp < i - 1)
            stgrp = st
            igrp = i
        rsa[pos] = igrp
        sa[i] = pos
    grpstart[igrp] = (igrp < size - 1 or size == 0)
    while grpstart.index(True) < size:
        # assert step <= size
        nextgr = grpstart.index(True)
        while nextgr < size:
            igrp = nextgr
            nextgr = grpstart.index(True, igrp + 1)
            glist = []
            for ig in range(igrp, nextgr):
                pos = sa[ig]
                if rsa[pos] != igrp:
                    break
                newgr = rsa[pos + step] if pos + step < size else -1
                glist.append((newgr, pos))
            glist.sort()
            for ig, g in groupby(glist, key=itemgetter(0)):
                g = [x[1] for x in g]
                sa[igrp:igrp + len(g)] = g
                grpstart[igrp] = (len(g) > 1)
                for pos in g:
                    rsa[pos] = igrp
                igrp += len(g)
        step *= 2
    del grpstart
    del rsa
    return np.array(sa)

@njit()
def search_suffix_array(searchstring, search_db, sa):
    lp = len(searchstring)
    n = len(sa)
    l = 0
    r = n
    while l < r:
        mid = int((l + r) / 2)
        a = sa[mid]
        if searchstring > search_db[a: a + lp]:
            l = mid + 1
        else:
            r = mid
    s = l
    r = n
    while l < r:
        mid = int((l + r) / 2)
        a = sa[mid]
        if searchstring < search_db[a: a + lp]:
            r = mid
        else:
            l = mid + 1
    return [sa[i] for i in range(s, r)]


def get_codons():
    bases = 'TCAG'
    return [a + b + c for a in bases for b in bases for c in bases]


def invert_codon_table(ct):
    """
    input: Codon table as dictionary with codons as keys and amino acids as values
    output: Dictionary with amino acids as keys and codons as values
    """
    inv_codon_table = clt.defaultdict(list)
    for k, v in ct.items():
        inv_codon_table[v] = inv_codon_table.get(v, [])
        inv_codon_table[v].append(k)
    return inv_codon_table


def get_mass_substitution_dict():
    """
    output: Dictionary of amino acid (key) mass differences (value)
    """
    aa_mass = {"G": 57.02147, "A": 71.03712, "S": 87.03203, "P": 97.05277, "V": 99.06842, "T": 101.04768,
               "I": 113.08407, "L": 113.08407, "N": 114.04293, "D": 115.02695, "Q": 128.05858, "K": 128.09497,
               "E": 129.0426, "M": 131.04049, "H": 137.05891, "F": 147.06842, "R": 156.10112, "C": 160.030654,  # CamCys
               "Y": 163.0633, "W": 186.07932, "*": 71.03712}

    subs_dict_complete = {i + ' to ' + j: aa_mass[j] - aa_mass[i] for i in aa_mass for j in aa_mass if i != j}
    del subs_dict_complete['L to I']
    del subs_dict_complete['I to L']

    subs_dict_complete = {k: v for k, v in subs_dict_complete.items() if k[-1] != 'L'}
    subs_dict = clt.defaultdict(int)
    for k, v in subs_dict_complete.items():  # unifies I and L
        if k[-1] == 'I':
            subs_dict[k + '/L'] = v
        else:
            subs_dict[k] = v

    return subs_dict


def codonify(seq):
    """
    input: a nucleotide sequence (not necessarily a string)
    output: a list of codons
    """
    seq = str(seq)
    l = len(seq)
    return [seq[i:i + 3] for i in range(0, l, 3)]


def is_gene(record, needs_stop_end=True):
    if len(record.seq) % 3 != 0:
        return False
    if not record.seq[:3] in {'ATG', 'GTG', 'TTG', 'ATT', 'CTG'}:
        return False
    if record.seq[-3:].translate() != '*' and needs_stop_end:
        return False
    return True


def c_term_probability(modified_sequence):
    """
    Returns the probability that C term AA was modified.
    """
    if modified_sequence[-1] == ')':
        return float(modified_sequence[:-1].split('(')[-1])
    else:
        return 0.0


def n_term_probability(modified_sequence):
    """
    Returns the probability that C term AA was modified.
    """
    if modified_sequence[1] == '(':
        return float(modified_sequence[2:].split(')')[0])
    else:
        return 0.0

@njit()
def is_prot_nterm(sequence, w_aa, suffix_array):
    """
    Does the peptide originate at the protein's N-term
    """
    for start in search_suffix_array(sequence, w_aa, suffix_array):
        if w_aa[start - 1] == '*':
            return True
        if w_aa[start - 2] == '*':
            return True
    return False

@njit()
def is_prot_cterm(sequence, w_aa, suffix_array):
    l = len(sequence)
    for start in search_suffix_array(sequence, w_aa, suffix_array):
        end = start + l
        if w_aa[end] == '*':
            return True
    return False


def refine_localization_probabilities(modified_seq, threshold=0.05):
    """
    returns the AAs that were possibly modified (with p > threshold).
    Input: modified sequence (a string of AA with p of each to contain modification: APKV(.7)ML(.3)L means
    that V was modified with p = .7 and L with p = .3)
    Output: A string with all candidate AAs separated by ';' (V;L).
    """
    modified_sites = [modified_seq[m.start() - 1] for m in re.finditer('\(', modified_seq)]
    weights = [float(i) for i in re.findall('\(([^\)]+)\)', modified_seq)]
    site_probabilities = {}
    for aa, weight in zip(modified_sites, weights):
        if aa in site_probabilities:
            site_probabilities[aa] += weight
        else:
            site_probabilities[aa] = weight
    return ";".join([k for k, v in site_probabilities.items() if v > threshold])


def read_allpeptide(filename, w_aa, sa, prot_bounds, prot_names, chunksize=10000, dp_columns=None):
    if dp_columns is None:
        dp_columns = [u'Raw file', u'Charge', u'm/z', u'Retention time', u'Sequence', u'Proteins', u'DP Base Sequence',
                      u'DP Mass Difference', u'DP Time Difference', u'DP PEP', u'DP Base Sequence', u'DP Probabilities',
                      u'DP Positional Probability', u'DP Decoy', u'DP Base Scan Number', u'DP Mod Scan Number']

    df_iter = pd.read_csv(filename, sep='\t', chunksize=chunksize, iterator=True, usecols=dp_columns)
    dp = pd.concat(chunk[pd.notnull(chunk['DP Mass Difference'])] for chunk in df_iter)
    dp.reset_index(drop=True, inplace=True)

    print('\tfinished reading... processing columns')

    dp['DPMD'] = dp['DP Mass Difference']
    dp['BASE_SCAN'] = dp['DP Base Scan Number']
    dp['MOD_SCAN'] = dp['DP Mod Scan Number']
    dp['DPAA_noterm'] = dp['DP Probabilities'].map(refine_localization_probabilities)
    dp['nterm'] = dp['DP Probabilities'].map(n_term_probability)  # p(N-term AA was substituted)
    dp['cterm'] = dp['DP Probabilities'].map(c_term_probability)
    # Does the peptide come from the N-term of the protein
    dp['prot_nterm'] = dp['DP Base Sequence'].map(lambda x: is_prot_nterm(x, w_aa, sa))
    dp['prot_cterm'] = dp['DP Base Sequence'].map(lambda x: is_prot_cterm(x, w_aa, sa))
    dp['proteins'] = dp['DP Base Sequence'].map(lambda x: find_proteins(x, prot_bounds, w_aa, sa, prot_names))
    dp['protein'] = dp['proteins'].map(lambda x: x.split(' ')[0] if len(x) > 0 else float('NaN'))
    dp['srt'] = False

    return dp


def find_peptide_start_position(protein_id, peptide, cds_aa):
    if protein_id != protein_id:  # test if protein_id is not NA or similar
        return -1
    db_cds = str(cds_aa[protein_id])
    # The minimum index is also the first occurrence in case the peptide occurs multiple times in the sequence
    # This means a potential SRT might be missed but this conservative approach is on purpose as we can not distinguish
    # which peptide we actuall see
    return db_cds.find(peptide)  # finds minimum index


def is_base_peptide_srt(protein_id, current_peptide, min_pos, wtlens):
    is_srt = False
    if protein_id != protein_id:
        return is_srt
    max_pos = min_pos + len(current_peptide)
    cds_length = wtlens[protein_id]
    if min_pos > cds_length or max_pos > cds_length:
        is_srt = True

    return is_srt


def is_dependent_peptide_srt(protein, position, wtlens):
    return wtlens[protein] - 1 < position


def find_proteins(base_seq, protein_bounds, w_aa, suffix_array, protein_names):
    """
    input: a peptide sequence (string)
    output: the names of proteins containing that sequence
    """
    tbr = " ".join([protein_names[i] for i in np.searchsorted(protein_bounds - 1,
                                                              search_suffix_array(base_seq, w_aa, suffix_array))])
    if tbr.strip(" ") == '':
        return ''
    else:
        return tbr


def fetch_codon_counts_in_peptide(peptide, cds_aa_dict, codons, prot_bounds, w_aa, sa, prot_names):
    """
    :param peptide: The peptide to search for in the proteome, for which the codons are identified
    :param cds_aa_dict: The proteome to search the peptide in
    :param codons: The codons to search for
    :return: DataFrame containing the counts of each codon in the sequence.
            Ambiguous peptides will be considered at all positions
    """

    possible_codons = np.zeros(len(codons))
    proteins = find_proteins(peptide, prot_bounds, w_aa, sa, prot_names)
    if proteins:
        proteins = proteins.split(" ")
    else:
        return possible_codons
    for p in proteins:
        if p in cds_aa_dict:
            s = cds_aa_dict[p].seq
            seq_i = s.translate().find(peptide)
            codon_seq = codonify(s)
            for i in range(seq_i, seq_i + len(peptide)):
                possible_codons[codons.index(codon_seq[i])] += 1

    return possible_codons


def count_observed_base_codons(peptide_df, cds_aa_dict, codons, prot_bounds, w_aa, sa, prot_names):
    """ Count all identified codons """
    codon_counts = np.zeros(len(codons))
    for peptide in peptide_df['DP Base Sequence']:
        codon_counts += fetch_codon_counts_in_peptide(peptide, cds_aa_dict, codons, prot_bounds, w_aa, sa, prot_names)

    return pd.DataFrame(data=codon_counts, index=codons, columns=['Count'])


def mark_dangerous_peptides(peptide_df, file_handle, aas, arguments):
    danger_mods = pd.read_csv(file_handle)
    peptide_df['danger'] = False
    for mod in danger_mods.iterrows():
        mod = mod[1]
        position = mod['position']
        site = mod['site']
        delta_m = mod['delta_m']

        mass_filter = (delta_m - (2 * arguments.tol) < peptide_df.DPMD) & \
                      (peptide_df.DPMD < delta_m + (2 * arguments.tol))

        term_filter = True
        if position == 'Protein N-term':
            term_filter = (peptide_df.nterm > arguments.n_term_prob_cutoff) & peptide_df.prot_nterm
        elif position == 'Protein C-term':
            term_filter = (peptide_df.cterm > arguments.c_term_prob_cutoff) & peptide_df.prot_cterm
        elif position == 'Any N-term':
            term_filter = peptide_df.nterm > arguments.n_term_prob_cutoff
        elif position == 'Any C-term':
            term_filter = peptide_df.cterm > arguments.c_term_prob_cutoff

        site_filter = True
        if site in aas:
            site_filter = peptide_df.DPAA_noterm.str.contains(site)

        peptide_df.loc[site_filter & term_filter & mass_filter, 'danger'] = True

    return peptide_df


def mark_substitutions(peptide_df, mass_diff_dict, tolerance, ppc):
    peptide_df['substitution'] = False
    for i in sorted(mass_diff_dict.keys()):
        delta_m = mass_diff_dict[i]
        original_aa = i[0]
        peptide_df.loc[(peptide_df.DPMD > delta_m - tolerance) & (peptide_df.DPMD < delta_m + tolerance) &
                       (peptide_df['DPAA_noterm'] == original_aa) & (peptide_df['DP Positional Probability'] > ppc) &
                       ~peptide_df['danger'], 'substitution'] = i
    return peptide_df


def hamming(s1, s2):
    return sum(a != b for a, b in zip(s1, s2))


def define_mispairing_mask(codons, aas, i_codon_table):
    """
    Create mask for mispairing. A binary dataframe indicating for each codon the AAs encoded by near-cognate codons.
    """
    mask = pd.DataFrame(data=False, index=codons, columns=list('ACDEFGHKLMNPQRSTVWY*'), dtype=float)
    for label in codons:
        near_cognates = np.array([hamming(i, label) == 1 for i in codons])
        reachable_aa = set(np.array(list(aas))[near_cognates])
        mask.loc[label] = [i in reachable_aa for i in 'ACDEFGHKLMNPQRSTVWY*']

    for label in mask.index:  # removes "near-cognates" that encodes the same AA
        for col in mask.columns:
            if label in i_codon_table[col]:
                mask.loc[label, col] = float('NaN')

    return mask


def fetch_codon(base_seq, modified_pos, cds_dict, prot_bounds, w_aa, sa, prot_names):
    """
    input: the original aa sequence of a peptide (base_seq),
            and the relative position of the modification.
    output: returns the list of all codons possibly associated
            with the substitution, presented as a string separated
            by white spaces.
    """
    possible_codons = []
    proteins = find_proteins(base_seq, prot_bounds, w_aa, sa, prot_names)
    if proteins:
        proteins = proteins.split(" ")
    else:
        return '_'
    for p in proteins:
        if p in cds_dict:
            s = cds_dict[p].seq
            seq_i = s.translate().find(base_seq)
            i = seq_i + modified_pos
            possible_codons.append(codonify(s)[i])
        else:
            possible_codons.append('_')
    return " ".join(possible_codons)


def fetch_best_codons(modified_seq, cds_dict, prot_bounds, w_aa, sa, prot_names):
    """
    input: a modified sequence, e.g. LQV(0.91)A(0.09)EK
    output: the list of codons associated with the most likely
            position
    """
    possible_sites = re.findall('\(([^\)]+)\)', modified_seq)
    best_site = np.argmax([float(i) for i in possible_sites])
    modified_pos_prime = [m.start() - 1 for m in re.finditer('\(', modified_seq)][best_site]
    modified_pos = len(re.sub('\(([^\)]+)\)', '', modified_seq[:modified_pos_prime]))
    base_seq = re.sub('\(([^\)]+)\)', '', modified_seq)
    return fetch_codon(base_seq, modified_pos, cds_dict, prot_bounds, w_aa, sa, prot_names)


def is_mispairing(codon, destination, mask):
    """
    Returns whether the substitution is mispairing or misloading, based on the
    near-cognate mask.
    """
    #codon = row['codon']
    #destination = row['destination']
    if pd.notnull(codon) and pd.notnull(destination):
        if (codon in mask.index) and destination:
            return mask.loc[codon, destination]
        else:
            return 0
    else:
        return float('NaN')


def find_substitution_position_local(modified_seq, protein, cds_dict):
    """
    returns the position of a substitutions relative to the start
    of the protein sequence
    """
    possible_sites = re.findall('\(([^\)]+)\)', modified_seq)
    best_site = np.argmax([float(i) for i in possible_sites])
    modified_pos_prime = [m.start() - 1 for m in re.finditer('\(', modified_seq)][best_site]
    modified_pos = len(re.sub('\(([^\)]+)\)', '', modified_seq[:modified_pos_prime]))
    base_seq = re.sub('\(([^\)]+)\)', '', modified_seq)
    s = cds_dict[protein].seq
    seq_i = s.translate().find(base_seq)
    i = seq_i + modified_pos
    return i


def find_positions_local(modified_seq, proteins, cds_dict):
    """
    returns the position of a substitutions relative to the start
    of the protein sequence, across all the codons
    """
    positions = []
    for prot in proteins.split(" "):
        positions.append(str(find_substitution_position_local(modified_seq, prot, cds_dict)))
    return " ".join(positions)


def create_modified_seq(modified_seq, destination):
    if type(destination) == str:
        possible_sites = re.findall('\(([^\)]+)\)', modified_seq)
        best_site = np.argmax([float(i) for i in possible_sites])
        modified_pos_prime = [m.start() - 1 for m in re.finditer('\(', modified_seq)][best_site]
        modified_pos = len(re.sub('\(([^\)]+)\)', '', modified_seq[: modified_pos_prime]))
        base_seq = re.sub('\(([^\)]+)\)', '', modified_seq)
        return base_seq[: modified_pos] + destination + base_seq[modified_pos + 1:]
    else:
        return ""


def find_homologous_peptide(peptide, is_srt, w_aa_ambiguous, sa_ambiguous):
    """
    Gets a peptide and returns whether it has homolegous in the genome.
    If so, that peptide is discarded.
    """
    if is_srt:
        return True
    if len(search_suffix_array('K' + peptide, w_aa_ambiguous, sa_ambiguous)) > 0:
        return False
    if len(search_suffix_array('R' + peptide, w_aa_ambiguous, sa_ambiguous)) > 0:
        return False
    if len(search_suffix_array('*' + peptide, w_aa_ambiguous, sa_ambiguous)) > 0:
        return False
    return True


def process_substitutions(peptide_df, proteome_string_ambiguous, suffix_array_ambiguous, mask, cds_dict, wtlen,
                          prot_bounds, w_aa, sa, prot_names, pos_prop_cutoff, fdr):
    subs = peptide_df[(peptide_df['substitution'] != False) | (peptide_df['srt'] != False)].copy()
    subs = subs[pd.notnull(subs['protein'])]  # mismatching files?

    subs['codons'] = 'NaN'
    subs.loc[subs['DP Positional Probability'] > pos_prop_cutoff, 'codons'] = \
        subs[subs['DP Positional Probability'] > pos_prop_cutoff]['DP Probabilities'].map(lambda x:
                                                                                          fetch_best_codons(x, cds_dict,
                                                                                                            prot_bounds,
                                                                                                            w_aa, sa,
                                                                                                            prot_names))
    subs['codon'] = subs['codons'].map(lambda x: x.split(' ')[0] if len(set(x.split(' '))) == 1 else float('NaN'))
    subs['destination'] = subs['substitution'].map(lambda x: x[-1] if x else False)
    subs['origin'] = subs['substitution'].map(lambda x: x[0] if x else False)
    subs['mispairing'] = subs.apply(lambda row: is_mispairing(row['codon'], row['destination'], mask), axis=1)

    subs['positions'] = subs.apply(lambda row: find_positions_local(row['DP Probabilities'], row['proteins'], cds_dict),
                                   axis=1)
    subs['position'] = subs['positions'].map(
        lambda x: int(x.split(' ')[0]) if len(set(x.split(' '))) == 1 else float('NaN'))
    subs['wt_len'] = subs['protein'].map(lambda x: wtlen[x])
    subs['modified_sequence'] = subs.apply(lambda row: create_modified_seq(row['DP Probabilities'], row['destination']),
                                           axis=1)
    subs['modified_sequence'] = subs['modified_sequence'].map(lambda x: x.replace('I', 'L'))
    subs = subs[subs.apply(lambda row: find_homologous_peptide(row['modified_sequence'], row['srt'],
                                                               proteome_string_ambiguous,
                                                               suffix_array_ambiguous), axis=1)]

    subs.sort_values('DP PEP', inplace=True)
    subs['decoy'] = pd.notnull(subs['DP Decoy'])
    # Understand this FDR protocol
    cut_off = np.max(np.where(np.array([i / float(j) for i, j in zip(subs['decoy'].cumsum(),
                                                                     range(1, len(subs) + 1))]) < fdr))
    subs = subs.iloc[:cut_off + 1]

    subs = subs[~subs.decoy]
    subs['srt'] = subs.apply(lambda row: row['srt'] or is_dependent_peptide_srt(row['protein'], row['position'], wtlen),
                             axis=1)

    return subs



from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from Bio import SeqIO, SeqUtils
from scipy import stats

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
import collections as clt
import seaborn as sns
import pandas as pd
import numpy as np

import os
import re
import argparse
from jinja2 import Environment, FileSystemLoader

#### FUNCTION DEFINITIONS AND GLOBAL VARIABLES (probably shouldn't have global variables)
aas = list('ACDEFGHIKLMNPQRSTVWY')
aas = [SeqUtils.seq3(aa) for aa in aas]
CODONTABLE = {'ATA': 'Ile', 'ATC': 'Ile', 'ATT': 'Ile', 'ATG': 'Met', 'ACA': 'Thr', 'ACC': 'Thr', 'ACG': 'Thr',
              'ACT': 'Thr', 'AAC': 'Asn', 'AAT': 'Asn', 'AAA': 'Lys', 'AAG': 'Lys', 'AGC': 'Ser', 'AGT': 'Ser',
              'AGA': 'Arg', 'AGG': 'Arg', 'CTA': 'Leu', 'CTC': 'Leu', 'CTG': 'Leu', 'CTT': 'Leu', 'CCA': 'Pro',
              'CCC': 'Pro', 'CCG': 'Pro', 'CCT': 'Pro', 'CAC': 'His', 'CAT': 'His', 'CAA': 'Gln', 'CAG': 'Gln',
              'CGA': 'Arg', 'CGC': 'Arg', 'CGG': 'Arg', 'CGT': 'Arg', 'GTA': 'Val', 'GTC': 'Val', 'GTG': 'Val',
              'GTT': 'Val', 'GCA': 'Ala', 'GCC': 'Ala', 'GCG': 'Ala', 'GCT': 'Ala', 'GAC': 'Asp', 'GAT': 'Asp',
              'GAA': 'Glu', 'GAG': 'Glu', 'GGA': 'Gly', 'GGC': 'Gly', 'GGG': 'Gly', 'GGT': 'Gly', 'TCA': 'Ser',
              'TCC': 'Ser', 'TCG': 'Ser', 'TCT': 'Ser', 'TTC': 'Phe', 'TTT': 'Phe', 'TTA': 'Leu', 'TTG': 'Leu',
              'TAC': 'Tyr', 'TAT': 'Tyr', 'TAA': 'Stop', 'TAG': 'Stop', 'TGC': 'Cys', 'TGT': 'Cys', 'TGA': 'Stop',
              'TGG': 'Trp'}
INV_CODON_TABLE = {'Ile': ['ATA', 'ATC', 'ATT'], 'Met': ['ATG'], 'Thr': ['ACA', 'ACC', 'ACG', 'ACT'],
                   'Asn': ['AAC', 'AAT'], 'Lys': ['AAA', 'AAG'], 'Ser': ['TCT', 'TCC', 'TCA', 'TCG', 'AGC', 'AGT'],
                   'Phe': ['TTT', 'TTC'], 'Arg': ['CGT', 'CGC', 'CGA', 'CGG', 'AGA', 'AGG'],
                   'Leu': ['CTA', 'CTC', 'CTG', 'CTT', 'TTA', 'TTG'], 'Pro': ['CCT', 'CCC', 'CCA', 'CCG'],
                   'His': ['CAT', 'CAC'], 'Gln': ['CAA', 'CAG'], 'Val': ['GTT', 'GTC', 'GTA', 'GTG'],
                   'Ala': ['GCT', 'GCC', 'GCA', 'GCG'], 'Asp': ['GAT', 'GAC'], 'Glu': ['GAA', 'GAG'],
                   'Gly': ['GGT', 'GGC', 'GGA', 'GGG'], 'Tyr': ['TAT', 'TAC'], 'Cys': ['TGC', 'TGT'], 'Trp': ['TGG'],
                   'Stop': ['TGA', 'TAG', 'TAA']}


def get_codons_by_aa():
    codons_by_aa = []
    for aa in aas:
        codons_by_aa = codons_by_aa + INV_CODON_TABLE[aa]

    return codons_by_aa


def unique(sequence):
    seen = set()
    return [x for x in sequence if not (x in seen or seen.add(x))]


def aa_limits(codon_list, ct):
    aas = []
    for c in codon_list:
        aas.append(ct[c])
    cc = dict(clt.Counter(aas))
    cc["Ala"] = 3.5
    cumsum = [0]
    for key, value in cc.items():
        cumsum.append(cumsum[-1] + value)
    cumsum.pop(0)
    del cumsum[-1]
    return cumsum


def read_codon_proteome(filename):
    proteome = list(SeqIO.parse(filename, "fasta"))

    codonified_proteome = clt.defaultdict(list)

    for protein in proteome:
        sequence = str(protein.seq).upper()
        codon_seq = [sequence[i: i+3] for i in range(0, len(sequence), 3)]
        codonified_proteome[protein.id] = codon_seq

    return codonified_proteome


def read_all_substitutions(folder, codon_proteome):
    files = os.listdir(folder)
    sub_files = sorted([os.path.join(folder, s) for s in files if 'errors' in s])
    substitution = pd.DataFrame(None, columns=['protein', 'destination', 'origin', 'codon', 'position_in_protein',
                                               'dataset'])
    for sf in sub_files:
        tmp = pd.read_csv(sf, sep=',', header=0,
                          usecols=['protein', 'destination', 'origin', 'codon', 'position_in_protein'])
        tmp['dataset'] = os.path.basename(sf).split('_')[0]
        substitution = pd.concat([substitution, tmp])

    substitution.dropna(axis=0, inplace=True)
    substitution['length'] = substitution['protein'].map(lambda x: len(codon_proteome[x]))
    substitution.query('length > 0', inplace=True)
    substitution['fractional_pos_in_prot'] = (substitution['position_in_protein'] / substitution[
        'length']) * 100

    return substitution


def read_peptide_counts(folder):
    files = os.listdir(folder)
    count_files = sorted([os.path.join(folder, s) for s in files if 'peptide_counts' in s])
    pep_count = pd.DataFrame(None, columns=['error_peptide', 'errorfree_peptide', 'total_peptide', 'dataset'])
    for cf in count_files:
        pxd_id = list(filter(lambda x: re.search(r'PXD', x), re.split('_|/', cf)))[0]
        tmp = pd.read_csv(cf, sep=',', header=0, index_col=0, usecols=['protein', 'error_peptide', 'errorfree_peptide',
                                                                       'total_peptide'])
        tmp['dataset'] = pxd_id
        pep_count = pd.concat([pep_count, tmp])

    return pep_count


def get_substitution_counts(subs_df):
    aa_substitution_count = pd.DataFrame(0, index=aas, columns=aas, dtype=int)

    codons = []
    for aa in aas:
        codons = codons + INV_CODON_TABLE[aa]

    codon_substitution_count = pd.DataFrame(0, index=codons, columns=aas, dtype=int)

    for row in subs_df.iterrows():
        r = row[1]
        o = SeqUtils.seq3(r['origin'])
        d_one = r['destination']
        if d_one == 'I/L':
            d_one = 'L'
        d = SeqUtils.seq3(d_one)
        aa_substitution_count.at[o, d] += 1
        codon_substitution_count.at[r['codon'], d] += 1

    return aa_substitution_count, codon_substitution_count


def read_substitutions_as_dict(folder, codon_proteome):
    files = os.listdir(folder)
    sub_files = sorted([os.path.join(folder, s) for s in files if 'errors' in s])
    substitution = {}

    for sf in sub_files:
        pxd_id = list(filter(lambda x: re.search(r'PXD', x), re.split('_|/', sf)))[0]
        tmp = pd.read_csv(sf, sep=',', header=0, usecols=['protein', 'destination', 'origin', 'codon',
                                                          'position_in_protein'])
        tmp.dropna(axis=0, inplace=True)
        tmp['length'] = tmp['protein'].map(lambda x: len(codon_proteome[x]))
        tmp.query('length > 0', inplace=True)
        tmp['fractional_pos_in_prot'] = (tmp['position_in_protein'] / tmp['length']) * 100

        substitution[pxd_id] = tmp

    return substitution


def read_codon_counts_as_dict(folder):
    files = os.listdir(folder)
    count_files = sorted([os.path.join(folder, s) for s in files if 'codon_counts' in s])
    codon_counts = {}
    for cf in count_files:
        pxd_id = list(filter(lambda x: re.search(r'PXD', x), re.split('_|/', cf)))[0]
        tmp = pd.read_csv(cf, sep=',', header=0, index_col=0, usecols=['Codon', 'total_count', 'error_count', 'detection_rate'])

        codon_counts[pxd_id] = tmp

    return codon_counts


def get_all_codon_count(cc_dict, subs_dict):
    all_codon_counts = pd.DataFrame(None, columns=['total_count', 'error_count', 'detection_rate'])
    for key in codon_count_dict.keys():
        cc = cc_dict[key]
        sub = subs_dict[key].groupby('codon').count()
        cc['subs'] = sub['protein']
        all_codon_counts = pd.concat([all_codon_counts, cc], axis=0)

    all_codon_counts['Codon'] = all_codon_counts.index
    all_codon_counts['AA'] = all_codon_counts['Codon'].map(lambda x: CODONTABLE[x])
    all_codon_counts['log_detection_rate'] = all_codon_counts['detection_rate'].map(np.log10)
    all_codon_counts.replace([np.inf, -np.inf], np.nan, inplace=True)
    all_codon_counts.dropna(subset=["log_detection_rate"], inplace=True)

    return all_codon_counts


def plot_substitution_identified(all_substitutions, peptide_count, cutoff=0.5, substitution_folder=None,
                                 filtered_folder=None, out_folder=None):

    n_subs = all_substitutions.groupby('dataset').count().loc[:, 'protein']
    n_peptides = peptide_count.groupby('dataset').sum().loc[:, 'total_peptide']

    x = np.array(np.log10(n_peptides))
    y = np.array(np.log10(n_subs))
    slope, intercept, r_value, p_value, std_err = stats.linregress(x, y)
    keep = np.where(((slope * x + intercept + cutoff) > y) & ((slope * x + intercept - cutoff) < y) & (y > 1))
    fig, axs = plt.subplots(1, 1, figsize=(6, 5))
    axs.scatter(x, y, color='0.25', alpha=0.3)
    axs.scatter(x[keep], y[keep], color='b', alpha=0.3)
    axs.set_xlabel(r'log$_{10}$[Identified peptides]', fontweight='bold')
    axs.set_ylabel(r'log$_{10}$[Identified substitutions]', fontweight='bold')
    axs.annotate(text=r"R$^2$ = {0:0.2f}".format(r_value), xy=(2, 3.2))
    axs.annotate(text=r"y = {0:.1f}x + {1:.1f}".format(slope, intercept), xy=(2, 3))
    axs.annotate(text=r"$\rho$ = {0:.4f}".format(p_value), xy=(2, 2.8))
    axs.axline(xy1=(min(x), slope * min(x) + intercept), slope=slope, color='0.25', linestyle='--')
    axs.axline(xy1=(min(x), slope * min(x) + intercept + cutoff), slope=slope, color='b', linestyle='--', alpha=0.3)
    axs.axline(xy1=(min(x), slope * min(x) + intercept - cutoff), slope=slope, color='b', linestyle='--', alpha=0.3)
    axs.axline(xy1=(min(x), 1), slope=0, color='b', linestyle='--', alpha=0.3)
    fig.tight_layout()
    plt.savefig(image_out_folder + '/global-1.png', format='png')

    # For HongKee: This part is tricky, of course the function should only plot. I took this whole routine from my
    # filtering function which I use to filter my datasets, only keeping what I deem useful.
    # Maybe it should be done separately which requires that we can ensure that the cutoff parameter is not changed
    # between the plot and the copying of files.
    if substitution_folder is not None and filtered_folder is not None:
        # keep only these datasets for fitting
        datasets = np.array(n_subs.index)[keep]
        for d in datasets:
            os.system(f'cp {substitution_folder}/{d}* filtered_folder')


def plot_substitution_pos_hist(all_substitutions, out_folder=None):
    fig, axs = plt.subplots(1, 1, figsize=(6, 5))
    # Ref: https://github.com/mwaskom/seaborn/issues/2652#issuecomment-916114304
    all_substitutions = all_substitutions.reset_index(drop=True)
    g = sns.histplot(data=all_substitutions, x='fractional_pos_in_prot', ax=axs, bins=100, color='0.25')
    g.set_xlabel('Percent Protein Length', fontweight='bold')
    g.set_ylabel('Detected Substitutions', fontweight='bold')
    fig.tight_layout()
    plt.savefig(image_out_folder + '/global-2.png', format='png')


def plot_heatmap(count_array, aabounds=None, colmap="bwr", log=False, include_total=False, out_folder=None):
    if include_total:
        columns = list(count_array.columns) + ["Total"]
    else:
        columns = list(count_array.columns)

    max_val = count_array.max().max()

    total = count_array.sum(axis=1)
    if log:
        total = np.log(total)
        count_array = np.log10(count_array)
        max_val = np.log10(max_val)
        count_array[np.isinf(count_array)] = float('NaN')

    img_data = pd.DataFrame(float('NaN'), index=count_array.index, columns=columns)
    img_data.at[count_array.index, count_array.columns] = count_array

    fig, ax = plt.subplots(figsize=(6.5, 10))
    plt.subplots_adjust(left=0.11, right=0.74, top=0.99, bottom=0.1)
    ima = ax.imshow(img_data, cmap=colmap, aspect='auto', vmax=max_val) #

    if include_total:
        img_data = pd.DataFrame(float('NaN'), index=count_array.index, columns=columns)
        img_data["Total"] = total
        imb = ax.imshow(img_data, cmap="Blues", aspect='auto', vmax=img_data.max().max())

    # Create colorbar
    if include_total:
        axins = inset_axes(ax, width="5%", height="45%", loc='lower left', bbox_to_anchor=(1.13, 0., 1, 1),
                           bbox_transform=ax.transAxes, borderpad=0)
        axins1 = inset_axes(ax, width="5%", height="45%", loc='upper left', bbox_to_anchor=(1.13, 0., 1, 1),
                            bbox_transform=ax.transAxes, borderpad=0)
    else:
        axins1 = inset_axes(ax, width="5%", height="45%", loc='center left', bbox_to_anchor=(1.13, 0., 1, 1),
                            bbox_transform=ax.transAxes, borderpad=0)

    if include_total:
        cbar = ax.figure.colorbar(imb, cax=axins, extend='both')
        plt.setp(cbar.ax.yaxis.get_ticklabels(), weight='bold')

    #ticks = np.linspace(0, np.ceil((10**max_val)/100), 5, endpoint=True)*100
    cbar2 = ax.figure.colorbar(ima, cax=axins1, extend='both')
    plt.setp(cbar2.ax.yaxis.get_ticklabels(), weight='bold')

    xlab = columns
    codons = count_array.index
    ylab = [c.replace('T', 'U') for c in count_array.index]

    # We want to show all ticks...
    ax.set_xticks(np.arange(len(xlab)))
    ax.set_yticks(np.arange(len(ylab)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(xlab, weight='bold')
    ax.set_yticklabels(ylab, weight='bold')

    # put x label on top
    ax.tick_params(top=False, bottom=True, labeltop=False, labelbottom=True)

    ax.set_xlabel("Destination Amino Acid", weight='bold')
    ax.set_ylabel("Original Codon", weight='bold')

    if aabounds is not None:
        ax.hlines(aabounds, *ax.get_xlim(), color='k')

        # second y-axis with AA labels
        ax1 = ax.twinx()
        ax1.set_ylim(ax.get_ylim())

        y2_tick_pos = [ax.get_ylim()[1]] + aabounds + [ax.get_ylim()[0]]
        y2_tick_pos = [sum(y2_tick_pos[i:i + 2]) / 2 for i in range(len(y2_tick_pos) - 2 + 1)]

        y2lab = unique([CODONTABLE[c] for c in codons])
        ax1.set_yticks(y2_tick_pos)
        ax1.set_yticklabels(y2lab, weight='bold', rotation=0)

        ax1.tick_params(top=False, bottom=True, labeltop=False, labelbottom=True)


        ax1.autoscale(False)

    ax.autoscale(False)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=90, ha="center", va="top", rotation_mode="default")
    plt.savefig(image_out_folder + '/global-3.png', format='png')


def plot_detection_rate(all_codon_counts, codons_by_aa, aabounds, out_folder=None):
    fig, ax = plt.subplots(1, 1, figsize=(6.5, 12))
    col1 = '#a6cee3'
    col2 = '#b2df8a'
    palette = {'Ala': col1, 'Cys': col2, 'Asp': col1, 'Glu': col2, 'Phe': col1, 'Gly': col2, 'His': col1, 'Ile': col2,
               'Lys': col1, 'Leu': col2, 'Met': col1, 'Asn': col2, 'Pro': col1, 'Gln': col2, 'Arg': col1, 'Ser': col2,
               'Thr': col1, 'Val': col2, 'Trp': col1, 'Tyr': col2}
    all_codon_counts.Codon = all_codon_counts.Codon.map(lambda x: x.replace('T', 'U'))
    codons = codons_by_aa
    codons_by_aa = [c.replace('T', 'U') for c in codons_by_aa]
    g = sns.boxplot(ax=ax, y="Codon", x="log_detection_rate", data=all_codon_counts, saturation=1,
                    order=codons_by_aa, hue='AA', dodge=False, fliersize=0, palette=palette)#'Set3')
    g = sns.stripplot(ax=ax, y="Codon", x="log_detection_rate", data=all_codon_counts, color='0.25',
                      order=codons_by_aa, alpha=0.1)
    g.set_xlabel('log$_{10}$[Error Detection Rate]', fontweight='bold')
    g.set_ylabel('Codon', fontweight='bold')
    plt.legend([], [], frameon=False)

    if aabounds is not None:
        ax.hlines(aabounds, *ax.get_xlim(), color='k')
        # second y-axis with AA labels
        ax1 = ax.twinx()
        ax1.set_ylim(ax.get_ylim())

        y2_tick_pos = [ax.get_ylim()[1]] + aabounds + [ax.get_ylim()[0]]
        y2_tick_pos = [sum(y2_tick_pos[i:i + 2]) / 2 for i in range(len(y2_tick_pos) - 2 + 1)]

        y2lab = unique([CODONTABLE[c] for c in codons])
        ax1.set_yticks(y2_tick_pos)
        ax1.set_yticklabels(y2lab, weight='bold', rotation=0)

        ax1.tick_params(top=False, bottom=True, labeltop=False, labelbottom=True)
        ax1.autoscale(False)
    ax.autoscale(False)

    fig.tight_layout()
    plt.savefig(image_out_folder + '/global-4.png', format='png')

def plot_substitutions_detected(subs_dict, out_folder=None):
    fig, ax = plt.subplots(1, 1, figsize=(6.5, 5))
    n_subs_detected = [len(df.index) for df in subs_dict.values()]
    g = sns.histplot(n_subs_detected, log_scale=True, color='0.25')
    g.set_ylabel('Datasets', fontweight='bold')
    g.set_xlabel('Detected Substitutions', fontweight='bold')
    fig.tight_layout()
    plt.savefig(image_out_folder + '/global-5.png', format='png')

parser = argparse.ArgumentParser()
parser.add_argument("-c", dest="cds_file", required=True,
                    help="cds filename", metavar="FILE")
parser.add_argument("-s", dest="substitution_folder", required=True,
                    help="substitution folder", metavar="FILE")

args = parser.parse_args()

print("\n\n\n")
print("--------------------------------------------------------------------")
print(f"cds_file: {args.cds_file}")
print(f"substitution_folder: {args.substitution_folder}")


#### FILENAMES AND STUFF
# substitution_folder = 'yeast_results_ionquant'
# cds_filename = '../fasta/s228c_orf_cds.fasta'
cds_filename = args.cds_file
substitution_folder = args.substitution_folder
out_folder = os.path.join(args.substitution_folder, 'report')
image_out_folder = os.path.join(out_folder, "images/global")

os.makedirs(image_out_folder, exist_ok=True)

peptide_count = read_peptide_counts(substitution_folder)
codon_prot = read_codon_proteome(cds_filename)
all_substitutions = read_all_substitutions(substitution_folder, codon_prot)
codon_count_dict = read_codon_counts_as_dict(substitution_folder)
substitution_dict = read_substitutions_as_dict(substitution_folder, codon_prot)

# plots require that all datasets have been processed
plot_substitution_identified(all_substitutions, peptide_count, cutoff=0.5, out_folder=out_folder)
plot_substitution_pos_hist(all_substitutions, out_folder=out_folder)

aa_sub_matrix, codon_sub_matrix = get_substitution_counts(all_substitutions)
aabounds = aa_limits(list(codon_sub_matrix.index), CODONTABLE)
plot_heatmap(codon_sub_matrix, aabounds=aabounds, colmap="Reds", log=True, include_total=False, out_folder=out_folder)

all_cc = get_all_codon_count(codon_count_dict, substitution_dict)
codons_by_aa = get_codons_by_aa()
plot_detection_rate(all_cc, codons_by_aa, aabounds, out_folder=out_folder)

plot_substitutions_detected(substitution_dict, out_folder=out_folder)

### HTML Report
lines = []
with open(out_folder + '/report_items.txt') as f:
    lines = f.readlines()

rawfiles = []
for line in lines:
    rawfiles.append({'name': line.rstrip("\n")})

# HTML Template creation
file_loader = FileSystemLoader('/workflow/templates')
env = Environment(loader=file_loader)
template = env.get_template('report.html')
output = template.render(pxds=rawfiles)

# Create results HTML file
html_file = open(out_folder + '/index.html', 'w')
html_file.write(output)
html_file.close()




from dataclasses import dataclass, field
import pandas as pd


@dataclass
class DataSet:
    substitution_file: str = ""
    count_file: str = ""

    substitutions: pd.DataFrame = field(default_factory=pd.DataFrame, init=True, repr=True)
    counts: pd.DataFrame = field(default_factory=pd.DataFrame, init=True, repr=True)

    # CONSTANTS
    ONE_LETTER_AA: list = field(default_factory=lambda: list("ACDEFGHKLMNPQRSTVWY"), repr=False)  # removed I!
    THREE_LETTER_AA: list = field(default_factory=lambda: ['Ala', 'Cys', 'Asp', 'Glu', 'Phe', 'Gly', 'His', 'Lys',
                                                           'Leu', 'Met', 'Asn', 'Pro', 'Gln', 'Arg', 'Ser', 'Thr',
                                                           'Val', 'Trp', 'Tyr'])
    ONE_TO_THREE: dict = field(default_factory=lambda: {'A': 'Ala', 'C': 'Cys', 'D': 'Asp', 'E': 'Glu', 'F': 'Phe',
                                                        'G': 'Gly', 'H': 'His', 'K': 'Lys', 'L': 'Leu', 'M': 'Met',
                                                        'N': 'Asn', 'P': 'Pro', 'Q': 'Gln', 'R': 'Arg', 'S': 'Ser',
                                                        'T': 'Thr', 'V': 'Val', 'W': 'Trp', 'Y': 'Tyr'})
    # I codons are translated as L
    CODONTABLE: dict = field(default_factory=lambda: {'ATA': 'L', 'ATC': 'L', 'ATT': 'L', 'ATG': 'M', 'ACA': 'T',
                                                      'ACC': 'T', 'ACG': 'T', 'ACT': 'T', 'AAC': 'N', 'AAT': 'N',
                                                      'AAA': 'K', 'AAG': 'K', 'AGC': 'S', 'AGT': 'S', 'AGA': 'R',
                                                      'AGG': 'R', 'CTA': 'L', 'CTC': 'L', 'CTG': 'L', 'CTT': 'L',
                                                      'CCA': 'P', 'CCC': 'P', 'CCG': 'P', 'CCT': 'P', 'CAC': 'H',
                                                      'CAT': 'H', 'CAA': 'Q', 'CAG': 'Q', 'CGA': 'R', 'CGC': 'R',
                                                      'CGG': 'R', 'CGT': 'R', 'GTA': 'V', 'GTC': 'V', 'GTG': 'V',
                                                      'GTT': 'V', 'GCA': 'A', 'GCC': 'A', 'GCG': 'A', 'GCT': 'A',
                                                      'GAC': 'D', 'GAT': 'D', 'GAA': 'E', 'GAG': 'E', 'GGA': 'G',
                                                      'GGC': 'G', 'GGG': 'G', 'GGT': 'G', 'TCA': 'S', 'TCC': 'S',
                                                      'TCG': 'S', 'TCT': 'S', 'TTC': 'F', 'TTT': 'F', 'TTA': 'L',
                                                      'TTG': 'L', 'TAC': 'Y', 'TAT': 'Y', 'TAA': 'Stop', 'TAG': 'Stop',
                                                      'TGC': 'C', 'TGT': 'C', 'TGA': 'Stop', 'TGG': 'W'}, repr=False)
    CODONS: list = field(default_factory=lambda: ["GCA", "GCC", "GCG", "GCT", "TGC", "TGT", "GAC", "GAT", "GAA",
                                                  "GAG", "TTT", "TTC", "GGT", "GGC", "GGA", "GGG", "CAT", "CAC",
                                                  "ATT", "ATC", "ATA", "AAA", "AAG", "TTA", "TTG", "CTT", "CTC",
                                                  "CTA", "CTG", "ATG", "AAT", "AAC", "CCT", "CCC", "CCA", "CCG",
                                                  "CAA", "CAG", "CGT", "CGC", "CGA", "CGG", "AGA", "AGG", "TCT",
                                                  "TCC", "TCA", "TCG", "AGT", "AGC", "ACT", "ACC", "ACA", "ACG",
                                                  "GTT", "GTC", "GTA", "GTG", "TGG", "TAT", "TAC"], repr=False)

    def __post_init__(self):
        # in some cases, there are multiple candidate codons even though the position is uniquely determined. Why?
        if self.count_file != "":
            df = pd.read_csv(self.count_file, index_col=1)
            df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
            self.set_codon_count(df=df)
        if self.substitution_file != "":
            subs = pd.read_csv(self.substitution_file)
            self.substitutions = self.create_substitution_matrix(subs, counts=self.counts)

    def create_substitution_matrix(self, df, counts):
        subs = df[['codon', 'destination', 'origin']]
        subs = subs.query("destination != 'False'")  # filter out any potential SRT (that are not also a substitution)
        subs = subs.dropna(axis=0)  # per row
        subs = subs.groupby(['codon', 'destination']).count()
        subs = subs.unstack(fill_value=0.0)
        # remove multilevel created by unstacking df
        subs.columns = subs.columns.droplevel(0)

        if '*' in subs.columns:
            subs.drop('*', axis=1, inplace=True)
        if 'I/L' in subs.columns:
            subs.rename(mapper={'I/L': 'L'}, axis=1, inplace=True)
        if 'L/I' in subs.columns:
            subs.rename(mapper={'L/I': 'L'}, axis=1, inplace=True)

        # add missing amino acid columns (in case some AA where not detected as substitution)
        if any(aa in subs.columns for aa in self.ONE_LETTER_AA):  # df uses one letter code
            subs.columns = [self.ONE_TO_THREE[aa] for aa in subs.columns]

        for aa in self.THREE_LETTER_AA:
            if aa not in subs:
                subs[aa] = 0.0
        # add missing codon rows (in case some codons had no substitution)
        for codon in self.CODONS:
            if codon not in subs.index:
                subs.loc[codon] = 0.0

        subs_per_codon = subs.sum(axis=1)
        # need loop since they are not in the same order, maybe sorting instead?
        for codon in counts.index:
            correct_incorporated = counts.at[codon, 'total_count'] - subs_per_codon.loc[codon]
            subs.at[codon, self.ONE_TO_THREE[self.CODONTABLE[codon]]] = int(correct_incorporated)

        return subs.astype('int32')

    def set_codon_count(self, df):
        if 'TAA' in df.index:
            df = df.drop('TAA')
        if 'TGA' in df.index:
            df = df.drop('TGA')
        if 'TAG' in df.index:
            df = df.drop('TAG')
        self.counts = df.astype('int32')

    def set_substitutions(self, df):
        self.substitutions = df

    def __add__(self, o):
        new_subs = self.substitutions + o.substitutions
        new_cc = self.counts + o.counts

        return DataSet(substitutions=new_subs, counts=new_cc)


if __name__ == "__main__":
    data = DataSet(substitution_file='~/repositories/darkproteome/output/cedric/pxd001714_errors.csv',
                   count_file='~/repositories/darkproteome/output/cedric/pxd001714_codon_count.csv')




import json

import argparse
import os

from jinja2 import Environment, FileSystemLoader

import spectrum_utils.iplot as sup
import spectrum_utils.spectrum as sus


def replacer(s, newstring, index, nofail=False):
    # raise an error if index is outside of the string
    if not nofail and index not in range(len(s)):
        raise ValueError("index outside given string")

    # if not erroring, but the index is still not in the correct range..
    if index < 0:  # add it to the beginning
        return newstring + s
    if index > len(s):  # add it to the end
        return s + newstring

    # insert the new string between "slices" of the original
    return s[:index] + newstring + s[index + 1:]


parser = argparse.ArgumentParser()
parser.add_argument("-b", dest="base_folder", required=True,
                    help="base folder name", metavar="FILE")
parser.add_argument("-f", dest="rawfile_folder", required=True,
                    help="rawfile folder name", metavar="FILE")
parser.add_argument("-o", dest="output_folder", required=True,
                    help="Output folder", metavar="FILE")

args = parser.parse_args()

print("\n")
print("--------------------------------------------------------------------")
print(f"base_folder: {args.base_folder}")
print(f"rawfile_folder: {args.rawfile_folder}")
print(f"Output folder: {args.output_folder}")

report_folder = os.path.join(args.output_folder, 'report')
out_folder = os.path.join(args.output_folder, 'report', args.rawfile_folder)
dataset_name = args.rawfile_folder
rawfile_folder = os.path.join(args.base_folder, args.rawfile_folder)

os.chdir(report_folder)

# Opening JSON file
f = open(dataset_name + '_peptides.js')

# returns JSON object as
# a dictionary
data = json.load(f)

# Closing file
f.close()

sus.static_modification('C', 57.021464)
peptides = []

# Iterating through the json
# list
for row in data:
    rawfile_name = row['rawfile_name']

    peak_list = row['peak_list']
    retention_time = row['retention_time']
    precursor_mz = row['precursor_mass']
    precursor_charge = row['precursor_charge']

    peptide, modifications = row['peptide'], {}
    spectrum_data = sus.MsmsSpectrum('Modified', precursor_mz, precursor_charge,
                                       [x[0] for x in peak_list],
                                       [x[1] for x in peak_list],
                                       peptide=peptide,
                                       retention_time=retention_time,
                                       modifications=modifications)

    org = row['origin']
    dest = row['destination']
    loc = row['mod_loc']

    peptide = replacer(peptide, f'[{org}\u279D{dest}]', loc)

    chart = (sup.spectrum(spectrum_data.annotate_peptide_fragments(0.5, 'Da', ion_types='by'))
             .properties(width=640, height=400, title=peptide))

    jsonSpec = chart.to_json()

    outputFile = dataset_name + '_' + rawfile_name + '_' + str(row['scan_number'])
    peptides.append({'rawfile_name': rawfile_name, 'outputFile': outputFile,
                     'scanNo': row['scan_number'], 'peptide': row['peptide'],
                     'jsonSpec': jsonSpec})


# Clean up
os.remove(f'{os.getcwd()}/{dataset_name}_peptides.js')

# HTML Template creation
file_loader = FileSystemLoader('/workflow/templates')
env = Environment(loader=file_loader)
template = env.get_template('substitution_report.html')
output = template.render(raw_folder=dataset_name, peptides=peptides)

html_file = open(f'{out_folder}.html', 'w')
html_file.write(output)
html_file.close()



import io
import time
import sys
from contextlib import redirect_stderr

import matplotlib.pyplot as plt
from tqdm import tqdm
from dataclasses import dataclass, field

import multiprocessing as mp
import numpy as np

from . import mtem
from . import data


# TODO change mtem to object that can be evaluated at a given parameter set in order to avoid passing model constants
#  and data through the mcmc. This also allows to keep things cleaner in general

@dataclass
class MCMCstate:
    alpha: np.array = field(default_factory=np.array)  # position penalty
    beta: np.array = field(default_factory=np.array)  # nucleotide mismatching
    log_likelihood: float = 0.0


@dataclass
class MCMC:
    model: mtem.MTEM

    c_alpha: np.array = field(init=False)  # position penalty
    c_beta: np.array = field(init=False)  # nucleotide mismatching

    p_alpha: np.array = field(init=False)  # position penalty
    p_beta: np.array = field(init=False)  # nucleotide mismatching

    alpha_acceptance_counter: int = 0
    beta_acceptance_counter: int = 0
    adaptive_width: int = 100
    a_proposal_width: float = 0.01
    b_proposal_width: float = 0.01

    trace: list = field(default_factory=list)

    current_log_likelihood: float = 0.0
    proposed_log_likelihood: float = 0.0

    POSITION_PENALTY = 'alpha'
    WOBBLE_PENALTY = 'beta'

    # Move the likelihood into a data class that 'knows' what type of data it is?
    # Probably better to move it into the model mtem class.
    # This would allow for a generalization of the MCMC. Necessary and worth the time?
    @staticmethod
    def log_likelihood(x, p):

        """
        :param x: data.frame of substitutions from codon (row) to AA (column)
        :param p: data.frame of estimated substitutions probabilities from codon (row) to AA (column)
        :return: log(likelihood) of x
        """
        log_p = np.log(np.array(p))
        ll = 0.0
        for dataset in x:
            d = np.array(dataset.substitutions)
            ll += np.sum(d * log_p)

        return ll

    @staticmethod
    def propose(x, sd):
        """
        :param x: vector of current position penalties
        :param sd: vector of proposal width
        :return: vector of proposed position penalties
        """
        y = np.random.normal(x, sd)
        return y

    def update_alpha(self):
        self.c_alpha = self.p_alpha
        self.current_log_likelihood = self.proposed_log_likelihood
        self.alpha_acceptance_counter += 1

    def update_beta(self):
        self.c_beta = self.p_beta
        self.current_log_likelihood = self.proposed_log_likelihood
        self.beta_acceptance_counter += 1

    def update_proposal_width(self):
        if self.alpha_acceptance_counter / self.adaptive_width < 0.25:
            self.a_proposal_width *= 0.8
        if self.alpha_acceptance_counter / self.adaptive_width > 0.35:
            self.a_proposal_width *= 1.2
        if self.beta_acceptance_counter / self.adaptive_width < 0.25:
            self.b_proposal_width *= 0.8
        if self.beta_acceptance_counter / self.adaptive_width > 0.35:
            self.b_proposal_width *= 1.2

        self.alpha_acceptance_counter = 0
        self.beta_acceptance_counter = 0

    def accept_parameter_set(self, log_parameter=False):
        """
        :return: boolean indicating acceptance/rejection of current state
        """

        ratio = (self.proposed_log_likelihood - self.current_log_likelihood)
        if log_parameter:
            ratio = ratio - np.sum(self.c_alpha - self.p_alpha)

        alpha = -np.random.exponential()

        return ratio > alpha

    def position_penalty_trace(self):
        nsamples = len(self.trace)
        position_trace = np.zeros(shape=(nsamples, self.trace[0].alpha.shape[0]))
        for i in range(nsamples):
            position_trace[i, :] = self.trace[i].alpha

        return np.exp(position_trace)

    def wobble_penalty_trace(self):
        nsamples = len(self.trace)
        wobble_trace = np.zeros(shape=(nsamples, self.trace[0].beta.shape[0]))
        for i in range(nsamples):
            wobble_trace[i, :] = self.trace[i].beta

        return wobble_trace

    def position_penalty_posterior(self, samples):
        nsamples = len(self.trace)
        posterior = np.zeros(shape=(samples, self.trace[0].alpha.shape[0]))
        for i, s in enumerate(range(nsamples - samples, nsamples)):
            posterior[i, :] = self.trace[s].alpha

        return np.exp(posterior)

    def wobble_penalty_posterior(self, samples):
        nsamples = len(self.trace)
        posterior = np.zeros(shape=(samples, self.trace[0].beta.shape[0]))
        for i, s in enumerate(range(nsamples - samples, nsamples)):
            posterior[i, :] = self.trace[s].beta

        return posterior

    def likelihood_trace(self):
        t_len = len(self.trace)
        ll_trace = [0] * t_len
        for t in range(t_len):
            ll_trace[t] = self.trace[t].log_likelihood

        return ll_trace

    def plot_posterior(self, which, samples):
        if which == self.POSITION_PENALTY:
            posterior = self.position_penalty_posterior(samples=samples)
            fig, axes = plt.subplots(1, 3)
            # TODO turn into loop
            axes[0].hist(posterior[:, 0])
            axes[0].set_xlabel('First Codon Position', weight='bold')
            axes[0].set_ylabel('Frequency', weight='bold')

            axes[1].hist(posterior[:, 1])
            axes[1].set_xlabel('First Codon Position', weight='bold')
            axes[1].set_ylabel('Frequency', weight='bold')

            axes[2].hist(posterior[:, 2])
            axes[2].set_xlabel('First Codon Position', weight='bold')
            axes[2].set_ylabel('Frequency', weight='bold')

    def prepare_dataset(self, unordered_data):
        reordered_data = []
        for d in unordered_data:
            d.substitutions = d.substitutions[self.model.THREE_LETTER_AA_NO_I]
            d.substitutions = d.substitutions.reindex(self.model.CODONS_NO_STOP)
            reordered_data.append(d)

        return reordered_data

    def run(self, samples, thinning, burnin, ic_alpha, ic_beta, data):
        data = self.prepare_dataset(data)

        iterations = (samples + burnin) * thinning
        self.trace = samples * [MCMCstate]

        self.c_alpha = ic_alpha
        self.c_beta = ic_beta
        c_beta = self.model.format_wobble_penalties(self.c_beta)
        p = self.model.calculate_all_substitution_probabilities(np.exp(self.c_alpha), c_beta, combine_ile_leu=True)
        self.current_log_likelihood = self.log_likelihood(data, p)
        print('Initial log(likelihood): {}\n'.format(self.current_log_likelihood))

        f = None
        if not sys.stdin.isatty():
            f = io.StringIO()

        for i in tqdm(range(0, iterations), file=sys.stdout):
            # Accept/Reject Position penalty
            self.p_alpha = self.propose(self.c_alpha, self.a_proposal_width)
            # forgetting this line caused a bug before. move it into the model itself?
            c_beta = self.model.format_wobble_penalties(self.c_beta)
            p = self.model.calculate_all_substitution_probabilities(np.exp(self.p_alpha), c_beta, combine_ile_leu=True)
            self.proposed_log_likelihood = self.log_likelihood(data, p)
            if self.accept_parameter_set(log_parameter=True):
                self.update_alpha()

            # Accept/Reject Nucleotide Mismatch penalty
            self.p_beta = self.propose(self.c_beta, self.b_proposal_width)
            p_beta = self.model.format_wobble_penalties(self.p_beta)
            p = self.model.calculate_all_substitution_probabilities(np.exp(self.c_alpha), p_beta, combine_ile_leu=True)
            self.proposed_log_likelihood = self.log_likelihood(data, p)
            if self.accept_parameter_set(log_parameter=False):
                self.update_beta()

            # ---- UPDATE TRACE ----
            if i / thinning >= burnin:
                if i % thinning == 0:
                    self.trace[int(i / thinning) - burnin] = MCMCstate(alpha=self.c_alpha, beta=self.c_beta,
                                                                       log_likelihood=self.current_log_likelihood)

            # ---- ADAPT PROPOSAL WIDTH ----
            if i != 0 and i % self.adaptive_width == 0:
                self.update_proposal_width()

            if f and i % 12 == 0:
                print(end ='\n', flush=True)

        print('Final log(likelihood): {}\n'.format(self.current_log_likelihood))



from random import choices
from datetime import date
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import argparse
import os
from gooey import Gooey, GooeyParser

from . import mtem
from . import data
from . import mcmc


def is_valid_file(p, arg):
    if not os.path.exists(arg):
        p.error("The file %s does not exist!" % arg)
    else:
        return arg


def prep_folder(arg):
    if not os.path.exists(arg):
        os.mkdir(arg)

    return arg


def run(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", dest="all_files", required=True, help="Folder with codon_count and error files",
                        metavar="FILE", type=lambda x: is_valid_file(parser, x))
    parser.add_argument("-r", dest="trna_count", required=True, help="tRNA count file", metavar="FILE",
                        type=lambda x: is_valid_file(parser, x))
    parser.add_argument("-o", dest="output_folder", required=True, help="Output folder", metavar="FILE",
                        type=lambda x: prep_folder(x))
    parser.add_argument("-s", dest="samples", required=True, type=int, help="Number of samples")
    parser.add_argument("-p", dest="post_samples", required=True, type=int, help="Posterior samples")
    parser.add_argument("-c", dest="cell_vol", required=False, help="Cell Volume", type=float, default=4.2e-17)
    parser.add_argument("-t", dest="thinning", required=False, help="Thinning", type=int, default=10)
    parser.add_argument("-b", dest="burnin", required=False, help="Burn-in", type=int, default=100)
    parser.add_argument("-nb", dest="num_bootsrap", required=False, help="Number of subsamples performed", type=int,
                        default=-1)
    parser.add_argument("-os", dest="out_suffix", required=False, help="suffix added to output files (default: date)",
                        type=str, default="")
    parser.add_argument("-a", dest="aggregate", required=False, type=str, default="n",
                        help="aggregate all datasets by summation (y,n) Default: No (n)")

    args = parser.parse_args(args)

    out_suffix = args.out_suffix
    if out_suffix == "":
        out_suffix = str(date.today())

    aggregate = False if args.aggregate == 'n' else 'True'
    print("\n\n\n")
    print("--------------------------------------------------------------------")
    print(f"Folder with codon_count and error files: {args.all_files}")
    print(f"Aggregate datasets: {aggregate}")
    print(f"tRNA count file: {args.trna_count}")
    print(f"Output folder: {args.output_folder}")
    print(f"Suffix: {out_suffix}")
    print(f"Number of samples: {args.samples}")
    print(f"Posterior samples: {args.post_samples}")
    print(f"Cell Volume: {args.cell_vol}")
    print(f"Thinning: {args.thinning}")
    print(f"Burn-in: {args.burnin}")
    print(f"Number of bootstraps: {args.num_bootsrap}")
    print("--------------------------------------------------------------------")

    Path(args.output_folder).mkdir(parents=True, exist_ok=True)

    global data

    files = os.listdir(args.all_files)
    cc_files = sorted([os.path.join(args.all_files, s) for s in files if 'codon_counts' in s])
    sub_files = sorted([os.path.join(args.all_files, s) for s in files if 'errors' in s])

    if len(cc_files) != len(sub_files):
        raise AssertionError("Not the same number of substitution files and codon count files")

    data_list = len(cc_files) * [None]
    i = 0
    for cc, sub in zip(cc_files, sub_files):
        dataset = data.DataSet(substitution_file=sub, count_file=cc)
        data_list[i] = dataset
        i = i + 1
    print(f"\tFound {len(data_list)} datasets")

    if aggregate:
        data = data_list[0]
        for i in range(1, len(data_list)):
            data = data + data_list[i]
        data_list = [data]

    model = mtem.MTEM(anticodon_file=args.trna_count, cell_volume=float(args.cell_vol))

    fitting = mcmc.MCMC(model=model)
    ic_pp = np.random.normal(size=2)
    ic_wp = np.random.normal(size=16)

    if args.num_bootsrap > 0:
        pos_means_df = pd.DataFrame(0, columns=['Pos 1', 'Pos 2'], index=range(args.num_bootsrap))
        wobble_means_df = pd.DataFrame(0, columns=model.WC_NUC_PAIRS, index=range(args.num_bootsrap))
        for k in range(args.num_bootsrap):
            bootstrapped_data = choices(data_list, k=len(data_list))
            print(f"\tPerforming Bootstrap {k + 1}/{args.num_bootsrap}")
            fitting.run(int(args.samples), int(args.thinning), int(args.burnin), ic_pp, ic_wp, bootstrapped_data)

            log_lik_trace = fitting.likelihood_trace()
            plt.plot(log_lik_trace)
            plt.savefig(os.path.join(args.output_folder, f"lik_trace_{k + 1}_{out_suffix}.png"))
            plt.close()

            ppp = fitting.position_penalty_posterior(int(args.post_samples))
            pos_means_df.loc[k] = np.mean(ppp, axis=0)
            wpp = fitting.wobble_penalty_posterior(int(args.post_samples))
            wobble_means_df.loc[k] = np.mean(wpp, axis=0)

        pos_means_df.to_pickle(os.path.join(args.output_folder, f"position_bootstrap_{out_suffix}.pkl"))
        wobble_means_df.to_pickle(os.path.join(args.output_folder, f"wobble_bootstrap_{out_suffix}.pkl"))
        pos_means_df.to_csv(os.path.join(args.output_folder, f"position_bootstrap_{out_suffix}.csv"))
        wobble_means_df.to_csv(os.path.join(args.output_folder, f"wobble_bootstrap_{out_suffix}.csv"))

    else:
        fitting.run(int(args.samples), int(args.thinning), int(args.burnin), ic_pp, ic_wp, data_list)
        hist_bins = np.max([int(int(args.post_samples) / 20), 10])

        # TODO: this part is a flipping mess. Needs cleanup
        # save position information
        position_trace = fitting.position_penalty_trace()
        position_trace = pd.DataFrame(position_trace, columns=['Pos1', 'Pos2'])
        position_trace.to_csv(os.path.join(args.output_folder, f"position_trace_{out_suffix}.csv"))
        position_trace.to_pickle(os.path.join(args.output_folder, f"position_trace_{out_suffix}.pkl"))

        ppp = fitting.position_penalty_posterior(int(args.post_samples))
        ppp = pd.DataFrame(ppp, columns=['Pos1', 'Pos2'])
        ppp.to_pickle(os.path.join(args.output_folder, f"position_{out_suffix}.pkl"))
        ppp.hist(density=True, figsize=(7, 4), bins=hist_bins)
        plt.tight_layout()
        plt.savefig(os.path.join(args.output_folder, f"position_{out_suffix}.png"))
        plt.close()
        ppp_mean = np.mean(ppp, axis=0)
        print(ppp_mean)
        ppp_mean.to_csv(os.path.join(args.output_folder, f"position_{out_suffix}.csv"))

        # save wobble information
        wobble_trace = fitting.wobble_penalty_trace()
        wobble_trace = pd.DataFrame(wobble_trace, columns=model.WC_NUC_PAIRS)
        wobble_trace.to_csv(os.path.join(args.output_folder, f"wobble_trace_{out_suffix}.csv"))
        wobble_trace.to_pickle(os.path.join(args.output_folder, f"wobble_trace_{out_suffix}.pkl"))

        wpp = fitting.wobble_penalty_posterior(int(args.post_samples))
        wpp = pd.DataFrame(wpp, columns=model.WC_NUC_PAIRS)
        wpp.to_csv(os.path.join(args.output_folder, f"wobble_{out_suffix}.csv"))
        wpp.to_pickle(os.path.join(args.output_folder, f"wobble_{out_suffix}.pkl"))
        wpp.hist(density=True, figsize=(15, 15), bins=hist_bins)
        plt.tight_layout()
        plt.savefig(os.path.join(args.output_folder, f"wobble_{out_suffix}.png"))
        plt.close()
        wpp_mean = np.mean(wpp, axis=0)
        wpp_mean.to_csv(os.path.join(args.output_folder, f"wobble_{out_suffix}.csv"))
        print(wpp_mean)

        # save likelihood information
        log_lik_trace = fitting.likelihood_trace()
        plt.plot(log_lik_trace)
        plt.tight_layout()
        plt.savefig(os.path.join(args.output_folder, f"lik_trace_{out_suffix}.png"))
        plt.close()
        log_lik_trace = pd.DataFrame(log_lik_trace, columns=['LogLikelihood'])
        log_lik_trace.to_csv(os.path.join(args.output_folder, f"lik_trace_{out_suffix}.csv"))
        log_lik_trace.to_pickle(os.path.join(args.output_folder, f"lik_trace_{out_suffix}.pkl"))

        # calculate substitution probabilities
        wp = model.format_wobble_penalties(np.array(wpp_mean))
        sub_probs = model.calculate_all_substitution_probabilities(np.array(ppp_mean), wp)
        sub_probs.to_csv(os.path.join(args.output_folder, f'fitted_substitution_probabilities_{out_suffix}.csv'))

        # store fitted energies
        df = pd.DataFrame(index=model.CODONS, columns=model.CODONS)
        for codon in model.CODONS:
            df[codon] = model.calculate_detection_parameter(model.CODONS, codon, ppp_mean, wp)
        df.to_csv(os.path.join(args.output_folder, f"fitted_energies_{out_suffix}.csv"))


from wx.core import wx
is_dark = wx.SystemSettings.GetAppearance().IsUsingDarkBackground()

print(is_dark)

bg_color = '#343434' if is_dark else '#eeeeee'
wbg_color = '#262626' if is_dark else '#ffffff'
fg_color = '#eeeeee' if is_dark else '#000000'

item_default = {
    'error_color': '#ea7878',
    'label_color': fg_color,
    'help_color': fg_color,
    'full_width': False,
    'validator': {
        'type': 'local',
        'test': 'lambda x: True',
        'message': ''
    },
    'external_validator': {
        'cmd': '',
    }
}

@Gooey(program_name='mTEL',
       default_size=(610, 700),
       terminal_panel_color=bg_color,
       terminal_font_color=fg_color,
       body_bg_color=bg_color,
       header_bg_color=wbg_color,
       footer_bg_color=bg_color,
       sidebar_bg_color=bg_color,)
def main(args):
    parser = GooeyParser(
        description="mTEL uses observed translation errors \nto estimate a multinomial translation error landscape.")

    for group in parser.parser._action_groups:
        group.gooey_options = {'label_color': fg_color, 'description_color': fg_color}

    parser.add_argument("-f", dest="all_files", required=True, help="Folder with codon_count and error files",
                        metavar="FILE", type=lambda x: is_valid_file(parser, x), widget="DirChooser", gooey_options=item_default)
    parser.add_argument("-r", dest="trna_count", required=True, help="tRNA count file", metavar="FILE",
                        type=lambda x: is_valid_file(parser, x), widget="FileChooser", gooey_options=item_default)
    parser.add_argument("-o", dest="output_folder", required=True, help="Output folder", metavar="FILE",
                        type=lambda x: prep_folder(x), widget="DirChooser", gooey_options=item_default)
    parser.add_argument("-s", dest="samples", required=True, type=int, help="Number of samples", gooey_options=item_default)
    parser.add_argument("-p", dest="post_samples", required=True, type=int, help="Posterior samples", gooey_options=item_default)
    parser.add_argument("-c", dest="cell_vol", required=False, help="Cell Volume", type=float, default=4.2e-17, gooey_options=item_default)
    parser.add_argument("-t", dest="thinning", required=False, help="Thinning", type=int, default=10, gooey_options=item_default)
    parser.add_argument("-b", dest="burnin", required=False, help="Burn-in", type=int, default=100, gooey_options=item_default)
    parser.add_argument("-nb", dest="num_bootsrap", required=False, help="Number of subsamples performed", type=int,
                        default=-1, gooey_options=item_default)
    parser.add_argument("-os", dest="out_suffix", required=False, help="suffix added to output files (default: date)",
                        type=str, default="", gooey_options=item_default)
    parser.add_argument("-a", dest="aggregate", required=False, type=str, default="n",
                        help="aggregate all datasets by summation (y,n) Default: No (n)", gooey_options=item_default)

    args = parser.parse_args()
    run(args)


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



import os
import time
import json


def load_json(filename):
    with open(filename, 'r') as f:
        data = json.load(f)
    return data['list']


def pride_filter(dataset):
    instrument_names = dataset['instrumentNames']
    return_value = False
    if len(instrument_names) == 1:
        return_value = 'LTQ Orbitrap' in instrument_names[0]

    return return_value


dict_list = load_json('pride/s_cerevisiae.json') # load json created with prideCrawler.py
filtered_pride_ids = [dataset['accession'] for dataset in dict_list] # get pride IDs, filter if needed, only basic filter for testing implemented but any filter can be build here
filtered_ftp_url = [dataset['_links']['datasetFtpUrl']['href'] for dataset in dict_list] # get needed ftp link

# download files and store them in correct directory. Each file will be placed in a directory named after the PRIDE id
for url, id in zip(filtered_ftp_url, filtered_pride_ids):
    if not os.path.isdir(id):
        os.system(f'mkdir {id}')
    os.system(f'cd {id}')
    os.system(f'wget "{url}/*.raw"')
    os.system('cd ..')
    time.sleep(1)



import sys

import pytest

from mTEL import mTEL
from eTEL import eTEL


def test_eTEL():
    eTEL.run(['-f', 'tests/resources/s228c_orf_cds.fasta', '-psm', 'tests/resources/psm.tsv', '-o', 'tests/output', '-decoy', 'rev_', '-p', 'substitutions', '-tol', '0.005', '-TEST', '-1'])
    assert True


def test_mTEL():
    mTEL.run(['-f', 'tests/resources/results_ionquant2', '-r', 'tests/resources/tRNA_count/yeast_tRNA_count.csv', '-o', 'tests/output', '-s', '250', '-p', '100', '-c', '4.2e-17', '-t', '10', '-b', '100', '-nb', '-1', '-a', 'n'])
    assert True



from gooey import Gooey, GooeyParser

from deTEL.eTEL.eTEL import is_valid_file, prep_folder

if __name__ == "__main__":
    from mTEL import mTEL
    from eTEL import eTEL
    # print('main')
else:
    from .mTEL import mTEL
    from .eTEL import eTEL
    # print('no_main')


from wx.core import wx
is_dark = wx.SystemSettings.GetAppearance().IsUsingDarkBackground()

bg_color = '#343434' if is_dark else '#eeeeee'
wbg_color = '#262626' if is_dark else '#ffffff'
fg_color = '#eeeeee' if is_dark else '#000000'

item_default = {
    'error_color': '#ea7878',
    'label_color': fg_color,
    'help_color': fg_color,
    'description_color': fg_color,
    'full_width': False,
    'external_validator': {
        'cmd': '',
    }
}

@Gooey(
    program_name="deTEL",
    program_description="detecting Translation Error Landscape",
    navigation="TABBED",
    default_size=(610, 700),
    terminal_panel_color=bg_color,
    terminal_font_color=fg_color,
    body_bg_color=bg_color,
    header_bg_color=wbg_color,
    footer_bg_color=bg_color,
    sidebar_bg_color=bg_color,
    progress_regex=r"(\d+)%"
)
def main():
    parser = GooeyParser()
    subs = parser.add_subparsers(help="commands", dest="command")

    eTEL_runner = subs.add_parser("eTEL", prog="eTEL")

    eTEL_args = eTEL_runner.add_argument_group("eTEL", description="eTEL detects the empirical translation error landscape.", gooey_options=item_default)
    eTEL_args.add_argument("-f", dest="fasta", required=True,
                           help="search database as codon fasta file", metavar="FILE",
                           type=lambda x: is_valid_file(parser, x), widget="FileChooser", gooey_options=item_default)
    eTEL_args.add_argument("-psm", dest="psm", required=True,
                           help="psm.tsv produced by philisopher", metavar="FILE", widget="FileChooser", gooey_options=item_default)
    eTEL_args.add_argument("-o", dest="output_folder", required=True,
                           help="Output folder", metavar="FILE",
                           type=lambda x: prep_folder(x), widget="DirChooser", gooey_options=item_default)
    eTEL_args.add_argument("-decoy", dest="decoy", required=False,
                           help="decoy prefix (default: rev_)", default="rev_", gooey_options=item_default)
    eTEL_args.add_argument("-p", dest="prefix", required=False,
                           help="prefix for experiment naming", default='substitutions', gooey_options=item_default)
    eTEL_args.add_argument("-tol", dest="tol", required=False,
                           help="m/z tolerance, used to filter DP–BP couples that resemble substitutions "
                                "and exclude pairs that resemble known PTM", default=0.005, gooey_options=item_default)
    eTEL_args.add_argument("-TEST", dest="test", required=False,
                           help="For testing only: use only <X> files", default=-1, gooey_options=item_default)

    mTEL_runner = subs.add_parser("mTEL", prog="mTEL")
    mTEL_args = mTEL_runner.add_argument_group("mTEL", description="mTEL uses observed translation errors \nto estimate a multinomial translation error landscape.", gooey_options=item_default)

    mTEL_args.add_argument("-f", dest="all_files", required=True, help="Folder with codon_count and error files",
                        metavar="FILE", type=lambda x: is_valid_file(parser, x), widget="DirChooser",
                        # default="tests/resources/results_ionquant2",
                           gooey_options=item_default)
    mTEL_args.add_argument("-r", dest="trna_count", required=True, help="tRNA count file", metavar="FILE",
                           # default="tests/resources/tRNA_count/yeast_tRNA_count.csv",
                        type=lambda x: is_valid_file(parser, x), widget="FileChooser", gooey_options=item_default)
    mTEL_args.add_argument("-o", dest="output_folder", required=True, help="Output folder", metavar="FILE",
                           # default="tests/output",
                        type=lambda x: prep_folder(x), widget="DirChooser", gooey_options=item_default)
    mTEL_args.add_argument("-s", dest="samples", required=True, type=int, help="Number of samples",
                            # default="250",
                           gooey_options=item_default)
    mTEL_args.add_argument("-p", dest="post_samples", required=True, type=int, help="Posterior samples",
                           # default="100",
                           gooey_options=item_default)
    mTEL_args.add_argument("-c", dest="cell_vol", required=False, help="Cell Volume", type=float, default=4.2e-17, gooey_options=item_default)
    mTEL_args.add_argument("-t", dest="thinning", required=False, help="Thinning", type=int, default=10, gooey_options=item_default)
    mTEL_args.add_argument("-b", dest="burnin", required=False, help="Burn-in", type=int, default=100, gooey_options=item_default)
    mTEL_args.add_argument("-nb", dest="num_bootsrap", required=False, help="Number of subsamples performed", type=int,
                        default=-1, gooey_options=item_default)
    mTEL_args.add_argument("-os", dest="out_suffix", required=False, help="suffix added to output files (default: date)",
                        type=str, default="", gooey_options=item_default)
    mTEL_args.add_argument("-a", dest="aggregate", required=False, type=str, default="n",
                        help="aggregate all datasets by summation (y,n) Default: No (n)", gooey_options=item_default)

    args = parser.parse_args()
    run(args)
    # convert(args.input_file, args.output_file)


def run(args):
    if args.command == "eTEL":
        eTEL.run(args)
    elif args.command == "mTEL":
        mTEL.run(args)


import sys

mode = sys.argv[1:]
print(mode)
if len(mode) == 0:
    main()
    exit(1)
else:
    if mode[0] == 'mTEL':
        mTEL.run(sys.argv[2:])
    elif mode[0] == 'eTEL':
        eTEL.run(sys.argv[2:])
    else:
        print('Argument error. Argument length: ', len(mode))
        exit(1)



from datetime import date

import os
import re
import json
import time
import math

import pandas as pd
import requests


def pride_filter(dataset):
    instrument_names = dataset['instruments']
    return_value = False
    if len(instrument_names) == 1:
        return_value = 'LTQ Orbitrap' in instrument_names[0]

    return return_value


# updated to work with new PRIDE API
def get_species_sets(species, outfile=None, removeMultiSpecies=False):
    time.sleep(0.5)
    url = f"https://www.ebi.ac.uk/pride/ws/archive/v2/search/projects?filter=organisms%3D%3D{species}"
    r = requests.get(url)
    if r.status_code != 200:
        raise ValueError('Wrong responds from website: ' + str(r.status_code))
        return None

    data = json.loads(r.content)
    saved_data = data['_embedded']['compactprojects']
    npages = data['page']['totalPages']
    print(f'number of pages: {npages}')
    print(f'finished page 1')
    pageurl = data['_links']['self']['href']
    for i in range(1, npages):
        current_url = re.sub('page=0', f'page={i}', pageurl, count=0, flags=0)
        r = requests.get(current_url)
        data = json.loads(r.content)
        saved_data += data['_embedded']['compactprojects']
        print(f'finished page {i+1}')

    if removeMultiSpecies:
        saved_data = [sd for sd in saved_data if len(sd['organisms']) == 1]

    if outfile is not None:
        with open(outfile, 'w', encoding='utf-8') as f:
            json.dump(saved_data, f, ensure_ascii=False, indent=4)

    return saved_data


def load_json(filename):
    with open(filename, 'r') as f:
        data = json.load(f)
    return data['list']


# e.g. find all datasets that contain 'sapiens' and store in json for future use
dict_list = get_species_sets('sapiens', 'h_sapiens.json',
                             removeMultiSpecies=True)

###### get number of potential datasets ##########
# Just examples if one is interested how many potential datasets are present per ...
#r = requests.get('https://www.ebi.ac.uk/pride/ws/archive/v2/stats/SUBMISSIONS_PER_ORGANISM')
#r = requests.get('https://www.ebi.ac.uk/pride/ws/archive/v2/stats/SUBMISSIONS_PER_DISEASES')
r = requests.get('https://www.ebi.ac.uk/pride/ws/archive/v2/stats/SUBMISSIONS_PER_INSTRUMENT')
data = json.loads(r.content)
df = pd.DataFrame(index=range(len(data)), columns=['Disease', 'nDatasets'])
for i, d in enumerate(data):
    df.iloc[i] = [d["key"], d["value"]]
df.to_csv('ndata_per_instrument_' + str(date.today()) + '.csv', index=False)

# in case you already created a json
#dict_list = load_json('h_sapiens.json')
# estimate storage space needed, likely an over estimate since only *.raw files will be downloaded but other formats are also considered 'RAW' by PRIDE
filtered_pride_ids = [dataset['accession'] for dataset in dict_list] # you could filter datasets here
nraw_files = []
for id in filtered_pride_ids:
    r = requests.get(f'https://www.ebi.ac.uk/pride/ws/archive/v2/files/byProject?accession={id}')
    if r.status_code != 200:
        continue
    data = json.loads(r.content)
    nraw_files.append(sum([d['fileSizeBytes'] * 1e-12 for d in data if d['fileCategory']['value'] == 'RAW']))
    time.sleep(0.5)

print(f'Estimated storage space required (TB): {math.ceil(sum(nraw_files))}')





import matplotlib
matplotlib.use('Agg')

import pandas as pd
import json

import argparse
import os

import ms_deisotope

parser = argparse.ArgumentParser()
parser.add_argument("-b", dest="base_folder", required=True,
                    help="base folder name", metavar="FILE")
parser.add_argument("-f", dest="rawfile_folder", required=True,
                    help="rawfile folder name", metavar="FILE")
parser.add_argument("-o", dest="output_folder", required=True,
                    help="Output folder", metavar="FILE")

args = parser.parse_args()

print("\n")
print("--------------------------------------------------------------------")
print(f"base_folder: {args.base_folder}")
print(f"rawfile_folder: {args.rawfile_folder}")
print(f"Output folder: {args.output_folder}")

report_folder = os.path.join(args.output_folder, 'report')
out_folder = os.path.join(args.output_folder, 'report', args.rawfile_folder)
dataset_name = args.rawfile_folder
rawfile_folder = os.path.join(args.base_folder, args.rawfile_folder)
substitution_error_file = os.path.join(args.output_folder, args.rawfile_folder + '_substitution_errors.csv')

peptides = []

psm = pd.read_csv(substitution_error_file)
os.chdir(report_folder)

groupedByRawfile = psm.groupby("Raw_file")
for name, group in groupedByRawfile:
    rawfile_name = str(name)
    print('\n')
    print(rawfile_name + '.raw')
    raw_file = os.path.join(args.base_folder, args.rawfile_folder, rawfile_name + '.raw')
    reader = ms_deisotope.MSFileLoader(raw_file)

    for index, row in group.iterrows():
        #         print(row)
        pep = {'rawfile_name': rawfile_name, 'scan_number': row.SCAN_NUM,
               'precursor_mass': row.calc_neutral_pep_mass, 'precursor_charge': row.charge,
               'retention_time': row.retention_time, 'origin': row.origin, 'destination': row.destination,
               'peptide': row.modified_peptide, 'best_locs': row.best_locs,
               'mod_loc': row.mod_loc, 'peptide_length': row.peptideLength}
        print(pep)

        scan = reader.get_scan_by_id(row.SCAN_NUM)
        scan.pick_peaks()
        peak_list = []
        for peak in scan.peak_set.peaks:
            peak_list.append([peak.mz, peak.intensity])

        pep['peak_list'] = peak_list
        peptides.append(pep)

f = open(dataset_name + "_peptides.js", "w")
f.write(json.dumps(peptides))
f.close()




import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt

import pandas as pd
import numpy as np
import os
import re
import argparse

fdr = 0.01

#### FUNCTION DEFINITIONS
def plot_codon_counts(cc, f):
    fig, ax1 = plt.subplots(1, 1, figsize=(10, 6))
    ax1.bar(cc.index, cc['total_count'])
    ax1.set_ylabel('Codon count', fontweight='bold')
    ax1.set_xlabel('Codons', fontweight='bold')
    ax1.set_xticklabels(cc.index, rotation=90)
    ax2 = ax1.twinx()
    ax2.plot(cc.index, cc['error_count'] / cc['total_count'], color='k',
             linestyle='dashed')
    ax2.plot(cc.index, cc['error_count'] / cc['total_count'], color='k', marker='o')
    ax2.set_ylabel('Error rate', fontweight='bold')
    plt.tight_layout()
    plt.savefig(f + '/1.png', format='png')


def plot_peptide_hist(pc, f):
    fig, ax1 = plt.subplots(1, 1, figsize=(7, 5))
    ax1.hist(pc['errorfree_peptide'], bins=100, histtype="stepfilled", alpha=0.8)
    ax1.set_xlabel('Error free peptides per protein', fontweight='bold')
    ax1.set_ylabel('Frequency', fontweight='bold')
    plt.tight_layout()
    plt.savefig(f + '/2.png', format='png')


def plot_hyperscore_distribution(psm, f):
    psm.sort_values('hyperscore', ascending=False, inplace=True)
    psm = psm.assign(is_decoy=psm['protein'].map(lambda x: re.search(r'rev_', x) is not None))
    ratio = np.array([i / float(j) for i, j in zip(psm['is_decoy'].cumsum(), range(1, len(psm.index) + 1))])

    if np.min(ratio) < fdr:
        cut_off = np.max(np.where(ratio < fdr))
        hyperscore_th = np.array(psm['hyperscore'])[cut_off]

    psm_real = psm.query('is_decoy == False')
    psm_decoy = psm.query('is_decoy == True')

    fig, ax1 = plt.subplots(1, 1, figsize=(7, 5))
    x = ax1.hist(psm_real['hyperscore'], bins=100, histtype="stepfilled", alpha=0.8, density=False, label='Identified')
    ax1.hist(psm_decoy['hyperscore'], bins=x[1], histtype="stepfilled", alpha=0.8, density=False, label='Decoys')
    ax1.set_xlabel('Hyperscore', fontweight='bold')
    ax1.set_ylabel('Density', fontweight='bold')
    if np.min(ratio) < fdr:
        ax1.axvline(hyperscore_th, color='k', linestyle='dashed')
    ax1.legend(frameon=False)
    plt.tight_layout()
    plt.savefig(f + '/3.png', format='png')


parser = argparse.ArgumentParser()
parser.add_argument("-f", dest="rawfile_folder", required=True,
                    help="rawfile folder name", metavar="FILE")
parser.add_argument("-b", dest="base_folder", required=True,
                    help="base folder name", metavar="FILE")
parser.add_argument("-s", dest="search_res_dir", required=True,
                    help="search result dir", metavar="FILE")
parser.add_argument("-o", dest="output_folder", required=True,
                    help="Output folder", metavar="FILE")

args = parser.parse_args()

print("\n\n\n")
print("--------------------------------------------------------------------")
print(f"base_folder: {args.base_folder}")
print(f"rawfile_folder: {args.rawfile_folder}")
print(f"search_res_dir_folder: {args.search_res_dir}")
print(f"Output folder: {args.output_folder}")


#### FILENAMES AND STUFF
# 'yeast_results_ionquant/PXD014119'
infolder = os.path.join(args.output_folder, args.rawfile_folder)
cf = '_'.join([infolder, 'codon_counts.csv'])
pf = '_'.join([infolder, 'peptide_counts.csv'])

tsv_folder = os.path.join(args.base_folder, args.rawfile_folder, args.search_res_dir, 'tsv')
out_folder = os.path.join(args.output_folder, 'report', args.rawfile_folder)

image_out_folder = os.path.join(args.output_folder, 'report/images', args.rawfile_folder)

os.makedirs(image_out_folder, exist_ok=True)

print(f"infolder: {infolder}")
print(f"cf: {cf}")
print(f"pf: {pf}")
print(f"tsv_folder: {tsv_folder}")

print("--------------------------------------------------------------------")

# after detection of substitution
codon_counts = pd.read_csv(cf, sep=',', header=0, index_col=0, usecols=['Codon', 'total_count', 'error_count'])
plot_codon_counts(codon_counts, image_out_folder)

peptide_counts = pd.read_csv(pf, sep=',', header=0, index_col=0)
plot_peptide_hist(peptide_counts, image_out_folder)

# after open search
tsv_files = [os.path.join(tsv_folder, f) for f in os.listdir(tsv_folder) if
             os.path.isfile(os.path.join(tsv_folder, f)) and ('.tsv' in f)]
psm_data = pd.DataFrame()
for tsvf in tsv_files:
    tmp_base = pd.read_csv(tsvf, sep='\t', header=0, usecols=['protein', 'retention_time', 'hyperscore'])
    psm_data = pd.concat([psm_data, tmp_base])
plot_hyperscore_distribution(psm_data, image_out_folder)






import io
from setuptools import setup


def README():
    with io.open('README.md', encoding='utf-8') as f:
        readme_lines = f.readlines()

    return ''.join(readme_lines)
README = README()  # NOQA

setup(
    name='deTELpy',
    version='0.1.6',
    description='Python package of the deTEL translation error detection pipeline from mass-spectrometry data',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://git.mpi-cbg.de/atplab/detelpy',
    author='Cedric Landerer',
    author_email='landerer@mpi-cbg.de',
    license='BSD',
    packages=['deTEL', 'deTEL.eTEL', 'deTEL.mTEL', 'deTEL.utility'],
    include_package_data=True,
    install_requires=['pandas==1.3.5',
                      'numpy',
                      'dataclasses>=0.6',
                      'pathlib>=1.0.1',
                      'biopython>=1.81',
                      'numba>=0.57.1',
                      'scipy>=1.7.3',
                      'tqdm>=4.65.0',
                      'seaborn>=0.12.2',
                      'matplotlib>=3.5.3',
                      'gooey>=1.0.7',
                      'python-dateutil'
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
)



from dataclasses import dataclass, field
from functools import partial
from numba import jit

import pandas as pd
import numpy as np


@dataclass
class MTEM:
    # CLASS VARIABLES
    anticodon_file: str
    cell_volume: float

    # fixed parameters
    trna: pd.DataFrame = field(init=False, repr=False)

    # calculated parameters
    arrival_rate: pd.DataFrame = field(init=False, repr=False)
    ar: np.array = field(init=False, repr=False)
    arrival_probabilities: pd.DataFrame = field(init=False, repr=False)
    ap: np.array = field(init=False, repr=False)
    synonymous_trna: dict = field(default_factory=dict, init=False, repr=False)
    available_trna: list = field(default_factory=list, init=False, repr=False)

    # CONSTANTS
    NON_WC_NUC_PAIRS: list = field(default_factory=lambda: ['AA', 'AC', 'AG', 'CA', 'CC', 'CT', 'TC', 'TT', 'TG', 'GA',
                                                            'GT', 'GG'], repr=False)
    WC_NUC_PAIRS: list = field(default_factory=lambda: ['AA', 'AC', 'AG', 'CA', 'CC', 'CT', 'TC', 'TT', 'TG', 'GA',
                                                        'GT', 'GG', 'AT', 'TA', 'GC', 'CG'], repr=False)
#    NON_CONST_WC_NUC_PAIRS: list = field(default_factory=lambda: ['AA', 'AC', 'AG', 'CA', 'CC', 'CT', 'TC', 'TT', 'TG',
#                                                                  'GA', 'GG', 'AT', 'TA', 'GC', 'CG'], repr=False)
    CODONS: list = field(default_factory=lambda: ["GCA", "GCC", "GCG", "GCT", "TGC", "TGT", "GAC", "GAT", "GAA",
                                                  "GAG", "TTT", "TTC", "GGT", "GGC", "GGA", "GGG", "CAT", "CAC",
                                                  "ATT", "ATC", "ATA", "AAA", "AAG", "TTA", "TTG", "CTT", "CTC",
                                                  "CTA", "CTG", "ATG", "AAT", "AAC", "CCT", "CCC", "CCA", "CCG",
                                                  "CAA", "CAG", "CGT", "CGC", "CGA", "CGG", "AGA", "AGG", "TCT",
                                                  "TCC", "TCA", "TCG", "AGT", "AGC", "ACT", "ACC", "ACA", "ACG",
                                                  "GTT", "GTC", "GTA", "GTG", "TGG", "TAT", "TAC", "TAA", "TAG",
                                                  "TGA"], repr=False)
    CODONS_NO_STOP: list = field(default_factory=lambda: ["GCA", "GCC", "GCG", "GCT", "TGC", "TGT", "GAC", "GAT", "GAA",
                                                          "GAG", "TTT", "TTC", "GGT", "GGC", "GGA", "GGG", "CAT", "CAC",
                                                          "ATT", "ATC", "ATA", "AAA", "AAG", "TTA", "TTG", "CTT", "CTC",
                                                          "CTA", "CTG", "ATG", "AAT", "AAC", "CCT", "CCC", "CCA", "CCG",
                                                          "CAA", "CAG", "CGT", "CGC", "CGA", "CGG", "AGA", "AGG", "TCT",
                                                          "TCC", "TCA", "TCG", "AGT", "AGC", "ACT", "ACC", "ACA", "ACG",
                                                          "GTT", "GTC", "GTA", "GTG", "TGG", "TAT", "TAC"], repr=False)
    INV_CODON_TABLE_SPLIT: dict = field(default_factory=lambda: {'Ile': ['ATA', 'ATC', 'ATT'], 'Met': ['ATG'],
                                                                 'Thr': ['ACA', 'ACC', 'ACG', 'ACT'],
                                                                 'Asn': ['AAC', 'AAT'], 'Lys': ['AAA', 'AAG'],
                                                                 'Ser2': ['AGC', 'AGT'], 'Gln': ['CAA', 'CAG'],
                                                                 'Ser4': ['TCT', 'TCC', 'TCA', 'TCG'],
                                                                 'Phe': ['TTT', 'TTC'], 'Arg2': ['AGA', 'AGG'],
                                                                 'Arg4': ['CGT', 'CGC', 'CGA', 'CGG'],
                                                                 'Leu': ['CTA', 'CTC', 'CTG', 'CTT'],
                                                                 'Leu2': ['TTA', 'TTG'], 'His': ['CAT', 'CAC'],
                                                                 'Pro': ['CCT', 'CCC', 'CCA', 'CCG'],
                                                                 'Val': ['GTT', 'GTC', 'GTA', 'GTG'],
                                                                 'Ala': ['GCT', 'GCC', 'GCA', 'GCG'],
                                                                 'Asp': ['GAT', 'GAC'], 'Glu': ['GAA', 'GAG'],
                                                                 'Gly': ['GGT', 'GGC', 'GGA', 'GGG'],
                                                                 'Tyr': ['TAT', 'TAC'], 'Cys': ['TGC', 'TGT'],
                                                                 'Trp': ['TGG'], 'Stop': ['TGA', 'TAG', 'TAA']},
                                        repr=False)

    INV_CODON_TABLE: dict = field(default_factory=lambda: {'Ile': ['ATA', 'ATC', 'ATT'], 'Met': ['ATG'],
                                                           'Thr': ['ACA', 'ACC', 'ACG', 'ACT'],
                                                           'Asn': ['AAC', 'AAT'], 'Lys': ['AAA', 'AAG'],
                                                           'Ser': ['TCT', 'TCC', 'TCA', 'TCG', 'AGC', 'AGT'],
                                                           'Phe': ['TTT', 'TTC'], 'Glu': ['GAA', 'GAG'],
                                                           'Arg': ['CGT', 'CGC', 'CGA', 'CGG', 'AGA', 'AGG'],
                                                           'Leu': ['CTA', 'CTC', 'CTG', 'CTT', 'TTA', 'TTG'],
                                                           'Pro': ['CCT', 'CCC', 'CCA', 'CCG'],
                                                           'His': ['CAT', 'CAC'], 'Gln': ['CAA', 'CAG'],
                                                           'Val': ['GTT', 'GTC', 'GTA', 'GTG'], 'Trp': ['TGG'],
                                                           'Ala': ['GCT', 'GCC', 'GCA', 'GCG'],
                                                           'Asp': ['GAT', 'GAC'], 'Stop': ['TGA', 'TAG', 'TAA'],
                                                           'Gly': ['GGT', 'GGC', 'GGA', 'GGG'],
                                                           'Tyr': ['TAT', 'TAC'], 'Cys': ['TGC', 'TGT']}, repr=False)
    ONE_LETTER_AA: list = field(default_factory=lambda: list("ACDEFGHIKLMNPQRSTVWY"), repr=False)
    ONE_LETTER_AA_NO_I: list = field(default_factory=lambda: list("ACDEFGHKLMNPQRSTVWY"), repr=False)
    THREE_LETTER_AA: list = field(default_factory=lambda: ['Ala', 'Cys', 'Asp', 'Glu', 'Phe', 'Gly', 'His', 'Ile',
                                                           'Lys', 'Leu', 'Met', 'Asn', 'Pro', 'Gln', 'Arg', 'Ser',
                                                           'Thr', 'Val', 'Trp', 'Tyr'])
    THREE_LETTER_AA_NO_I: list = field(default_factory=lambda: ['Ala', 'Cys', 'Asp', 'Glu', 'Phe', 'Gly', 'His', 'Lys',
                                                                'Leu', 'Met', 'Asn', 'Pro', 'Gln', 'Arg', 'Ser', 'Thr',
                                                                'Val', 'Trp', 'Tyr'])

    CODONTABLE_SPLIT: dict = field(default_factory=lambda: {'ATA': 'Ile', 'ATC': 'Ile', 'ATT': 'Ile', 'ATG': 'Met',
                                                            'ACA': 'Thr', 'ACC': 'Thr', 'ACG': 'Thr', 'ACT': 'Thr',
                                                            'AAC': 'Asn', 'AAT': 'Asn', 'AAA': 'Lys', 'AAG': 'Lys',
                                                            'AGC': 'Ser2', 'AGT': 'Ser2', 'AGA': 'Arg2',
                                                            'AGG': 'Arg2', 'CTA': 'Leu', 'CTC': 'Leu',
                                                            'CTG': 'Leu', 'CTT': 'Leu', 'CCA': 'Pro', 'CCC': 'Pro',
                                                            'CCG': 'Pro', 'CCT': 'Pro', 'CAC': 'His', 'CAT': 'His',
                                                            'CAA': 'Gln', 'CAG': 'Gln', 'CGA': 'Arg', 'CGC': 'Arg',
                                                            'CGG': 'Arg', 'CGT': 'Arg', 'GTA': 'Val', 'GTC': 'Val',
                                                            'GTG': 'Val', 'GTT': 'Val', 'GCA': 'Ala', 'GCC': 'Ala',
                                                            'GCG': 'Ala', 'GCT': 'Ala', 'GAC': 'Asp', 'GAT': 'Asp',
                                                            'GAA': 'Glu', 'GAG': 'Glu', 'GGA': 'Gly', 'GGC': 'Gly',
                                                            'GGG': 'Gly', 'GGT': 'Gly', 'TCA': 'Ser4',
                                                            'TCC': 'Ser4', 'TCG': 'Ser4', 'TCT': 'Ser4',
                                                            'TTC': 'Phe', 'TTT': 'Phe', 'TTA': 'Leu2',
                                                            'TTG': 'Leu2', 'TAC': 'Tyr', 'TAT': 'Tyr',
                                                            'TAA': 'Stop', 'TAG': 'Stop', 'TGC': 'Cys',
                                                            'TGT': 'Cys', 'TGA': 'Stop', 'TGG': 'Trp',
                                                            'LAT': 'Ile'}, repr=False)
    CODONTABLE: dict = field(default_factory=lambda: {'ATA': 'Ile', 'ATC': 'Ile', 'ATT': 'Ile', 'ATG': 'Met',
                                                      'ACA': 'Thr', 'ACC': 'Thr', 'ACG': 'Thr', 'ACT': 'Thr',
                                                      'AAC': 'Asn', 'AAT': 'Asn', 'AAA': 'Lys', 'AAG': 'Lys',
                                                      'AGC': 'Ser', 'AGT': 'Ser', 'AGA': 'Arg', 'AGG': 'Arg',
                                                      'CTA': 'Leu', 'CTC': 'Leu', 'CTG': 'Leu', 'CTT': 'Leu',
                                                      'CCA': 'Pro', 'CCC': 'Pro', 'CCG': 'Pro', 'CCT': 'Pro',
                                                      'CAC': 'His', 'CAT': 'His', 'CAA': 'Gln', 'CAG': 'Gln',
                                                      'CGA': 'Arg', 'CGC': 'Arg', 'CGG': 'Arg', 'CGT': 'Arg',
                                                      'GTA': 'Val', 'GTC': 'Val', 'GTG': 'Val', 'GTT': 'Val',
                                                      'GCA': 'Ala', 'GCC': 'Ala', 'GCG': 'Ala', 'GCT': 'Ala',
                                                      'GAC': 'Asp', 'GAT': 'Asp', 'GAA': 'Glu', 'GAG': 'Glu',
                                                      'GGA': 'Gly', 'GGC': 'Gly', 'GGG': 'Gly', 'GGT': 'Gly',
                                                      'TCA': 'Ser', 'TCC': 'Ser', 'TCG': 'Ser', 'TCT': 'Ser',
                                                      'TTC': 'Phe', 'TTT': 'Phe', 'TTA': 'Leu', 'TTG': 'Leu',
                                                      'TAC': 'Tyr', 'TAT': 'Tyr', 'TAA': 'Stop', 'TAG': 'Stop',
                                                      'TGC': 'Cys', 'TGT': 'Cys', 'TGA': 'Stop', 'TGG': 'Trp',
                                                      'LAT': 'Ile'}, repr=False)

    WC: dict = field(default_factory=lambda: {"A": "T", "C": "G", "G": "C", "T": "A", "L": "A"}, repr=False)
    NUC: list = field(default_factory=lambda: list("ATCG"), repr=False)
    NUC_IDX: dict = field(default_factory=lambda: {"A": 0, "C": 1, "G": 2, "T": 3, "N": 4, "L": 5}, repr=False)

    REV_COMPL: dict = field(default_factory=dict, init=False, repr=False)

    # FUNCTIONS
    def __post_init__(self):
        self.REV_COMPL = {c: self.reverse_complement(c) for c in self.CODONS}

        self.codon_idx = {codon: self.THREE_LETTER_AA.index(self.CODONTABLE[codon]) for codon in self.CODONS_NO_STOP}

        self.trna = pd.read_csv(self.anticodon_file, delimiter=",", header=0)

        # drop STOP AC until we have actual energies for the RFs
        self.trna.index = self.trna['Anticodon']
        self.trna = self.trna.drop(['TTA', 'CTA', 'TCA'], axis=0)

        self.available_trna = list(self.trna['Anticodon'])

        self.arrival_rate = self.calculate_arrival_rates()
        self.arrival_probabilities = self.calculate_arrival_probabilities()

        aa_group = self.arrival_rate.groupby('AA')
        for aa in self.THREE_LETTER_AA:  # Add STOP codons once we have actual energies
            self.synonymous_trna[aa] = list(aa_group.get_group(aa)['Anticodon'])

        # these two are needed for the acceptance probability. Pre-cast them to numpy as this seems to take quite a
        # bit of the total computation time
        self.ap = np.array(self.arrival_probabilities['First'])
        self.ar = np.array(self.arrival_rate['ArrivalRate'])

    def format_wobble_penalties(self, wobble_param):
        """
        :param wobble_param: wobble parameters exculding Watson-Crick pairs (12)
        :return: Dictionary with wobble parameters, and Watson-Crick Pairs set to 1
        """

        ret = dict(zip(self.WC_NUC_PAIRS, wobble_param))
        return ret

    def reverse_complement(self, current_codon):
        return self.WC[current_codon[2]] + self.WC[current_codon[1]] + self.WC[current_codon[0]]

    def calculate_arrival_rates(self):
        # Parameters taken from Shah et. al 2013
        effect_length = 1.5e-8
        diffusion_coeff = 8.42e-11
        transition_time = (effect_length ** 2) / (6 * diffusion_coeff)  # transition time between locations
        n_loc = self.cell_volume / effect_length ** 3  # number of locations in a cell

        ar = pd.DataFrame()
        ar['Anticodon'] = self.available_trna
        ar.index = ar['Anticodon']
        ar['ArrivalRate'] = self.trna['Abundance'].map(lambda x: (x / n_loc) / transition_time)
        ar['AA'] = self.trna['Anticodon'].map(lambda x: self.CODONTABLE[self.REV_COMPL[x]])
        return ar

    def calculate_arrival_probabilities(self):
        denominator = self.arrival_rate['ArrivalRate'].sum()
        df_index = self.available_trna
        ap = pd.DataFrame()
        aa_rate = self.arrival_rate.groupby('AA').sum(numeric_only=True)

        ap['First'] = self.arrival_rate['ArrivalRate'].map(lambda x: x / denominator)
        ap['First_syn'] = self.arrival_rate.apply(lambda row: row['ArrivalRate'] / aa_rate.at[row['AA'], 'ArrivalRate'],
                                                  axis=1)
        for ac in self.available_trna:
            focal_ac = self.arrival_rate.at[ac, 'ArrivalRate']
            ap[ac] = self.arrival_rate['ArrivalRate'].map(lambda x: focal_ac / (focal_ac + x))
        ap.index = df_index
        return ap

    @staticmethod
    def calc_dr(ac, c, position_penalty, wobble_set):
        wp1 = wobble_set[c[0] + ac[2]]
        wp2 = wobble_set[c[1] + ac[1]]
        wp3 = wobble_set[c[2] + ac[0]]

        # position penalty for position three is fixed to 1
        return position_penalty[0] * wp1 + position_penalty[1] * wp2 + wp3

    def calculate_detection_parameter(self, trnas, c, position_penalty, wobble_set):
        dr = np.array([self.calc_dr(ac, c, position_penalty, wobble_set) for ac in trnas])

        return dr

    def calculate_asite_binding_probability(self, acodon, trnas, position_penalty=None, wobble_set=None):
        detection_rates = self.calculate_detection_parameter(trnas, acodon, position_penalty, wobble_set)
        delta_stability = np.exp(-detection_rates)  # stability of codon/anticodon interaction
        p_binding_a = delta_stability / np.sum(delta_stability)

        return p_binding_a

    def calculate_acceptance_probability(self, pb):
        nrow = pb.shape[0]
        not_p_b = 1 - pb
        return self.calculate_acceptance_probability_jit(pb, not_p_b, self.ap, self.ar, nrow)

    @staticmethod
    @jit(nopython=True)
    def calculate_acceptance_probability_jit(pb, not_p_b, ap, ar, n):
        p_acceptance = np.zeros(pb.shape)
        for ac1 in range(n):
            p_acceptance[ac1] = pb[ac1] * ap[ac1]
            x = np.delete(not_p_b, ac1)
            y = np.delete(ar, ac1)
            tmp = np.prod(np.power(x, y / ar[ac1]))
            p_acceptance[ac1] += (tmp * pb[ac1]) * (1 - ap[ac1])

        return p_acceptance / np.sum(p_acceptance)

    def calculate_substitution_probability(self, acodon, position_penalty, wobble_set):
        trnas = self.available_trna
        asite_binding_probability = self.calculate_asite_binding_probability(acodon, trnas, position_penalty,
                                                                             wobble_set)
        return self.calculate_acceptance_probability(asite_binding_probability)

    def calculate_all_substitution_probabilities(self, position_penalty, wobble_set, combine_ile_leu=False):
        np_sub_probs = np.zeros(shape=(len(self.CODONS_NO_STOP), len(self.available_trna)), dtype='float')

        for i, current_codon in enumerate(self.CODONS_NO_STOP):
            np_sub_probs[i] = self.calculate_substitution_probability(current_codon, position_penalty, wobble_set)

        # Convert codon to anticodon dataframe to codon to amino acid dataframe
        np_codon_to_aa = np.zeros(shape=(len(self.CODONS_NO_STOP), len(self.THREE_LETTER_AA)), dtype='float')
        for c, current_codon in enumerate(self.CODONS_NO_STOP):
            for ac, anticodon in enumerate(self.available_trna):
                trna = self.REV_COMPL[anticodon]
                aa = self.codon_idx[trna]  # amino acid of the attempting tRNA
                np_codon_to_aa[c, aa] += np_sub_probs[c, ac]

        codon_to_aa = pd.DataFrame(np_codon_to_aa, index=self.CODONS_NO_STOP, columns=self.THREE_LETTER_AA)
        if combine_ile_leu:
            codon_to_aa['Leu'] = codon_to_aa['Ile'] + codon_to_aa['Leu']
            codon_to_aa = codon_to_aa.drop('Ile', axis=1)
            codon_to_aa = codon_to_aa[self.THREE_LETTER_AA_NO_I]

        # Numerics (double check that this is not a bug) causes near 0 negative numbers.
        codon_to_aa[codon_to_aa <= 0] = 1e-20  # avoid zero since we want to take the log in the MCMC
        codon_to_aa = codon_to_aa.div(codon_to_aa.sum(axis=1), axis=0)

        return codon_to_aa

    def simulate_data(self, position_penalty, wobble_penalty):
        substitution_probabilities = self.calculate_all_substitution_probabilities(position_penalty, wobble_penalty,
                                                                                   combine_ile_leu=True)
        simulated_codon_counts = pd.DataFrame(0, index=substitution_probabilities.index,
                                              columns=substitution_probabilities.columns)
        ncodons = len(substitution_probabilities.index)
        n_codons_observed = pd.DataFrame(0, index=substitution_probabilities.index, columns=['Count'])
        n_codons_observed['Count'] = 10000 * np.random.randn(ncodons) + 45000
        for codon in substitution_probabilities.index:
            simulated_codon_counts.loc[codon:, ] = np.random.multinomial(n_codons_observed.at[codon, 'Count'],
                                                                         substitution_probabilities.loc[codon, :])
        return n_codons_observed.astype('int32'), simulated_codon_counts.astype('int32')

    def calculate_incorporation_probability(self, seq, position_penalty, wobble_penalty, include_fs=True):
        incorporation_df = pd.DataFrame()
        if include_fs:
            # default is full 10-mer
            is_start, is_start = False, False
            if len(seq) == 5:
                is_start = (seq[0:3] == 'ATG')
                is_end = not is_start
            # TODO implement FS model
        else:
            incorporation_df = self.calculate_substitution_probability(seq, position_penalty, wobble_penalty)

        return incorporation_df


'''
if __name__ == "__main__":
    model = MTEM(anticodon_file="parameter_files/ecoli_tRNA_count_SRR1232430.csv", cell_volume=0.6e-18)

    pp = np.array(pd.read_csv('output/pride_filtered_ecoli/position_2022-03-10.csv', index_col=0, header=0)).flatten()
    w = np.array(pd.read_csv('output/pride_filtered_ecoli/wobble_2022-03-10.csv', index_col=0, header=0)).flatten()
    wp = model.format_wobble_penalties(w)
    test = model.calculate_all_substitution_probabilities(pp, wp)

    test = pd.Series(model.calculate_asite_binding_probability('ATT', model.available_trna, pp, wp))
    test.index = model.available_trna

    df = pd.DataFrame(index=model.CODONS, columns=model.CODONS)
    for codon in model.CODONS:
        df[codon] = model.calculate_detection_parameter(model.CODONS, codon, pp, wp)

    df.to_csv('output/ecoli/fitted_energies.csv')
'''